#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

from src.pytrain.pdi.base3_component import ConsistComponent, RouteComponent, UnitBits
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum


class TestConsistComponent:
    def test_as_bytes_and_properties_single_forward(self):
        # flags: unit SINGLE (00), forward (bit 2 = 0), train_linked off, horn/dialog off, tmcc2 off, accessory off
        flags = 0b00000000
        c = ConsistComponent(tmcc_id=7, flags=flags)

        assert c.tmcc_id == 7
        assert c.flags == flags
        assert c.unit_type == UnitBits.SINGLE
        assert c.is_single is True
        assert c.is_head is False
        assert c.is_middle is False
        assert c.is_tail is False

        assert c.is_forward is True
        assert c.is_reverse is False

        assert c.is_train_linked is False
        assert c.is_horn_masked is False
        assert c.is_dialog_masked is False
        assert c.is_tmcc2 is False
        assert c.is_accessory is False

        # as_bytes: [flags][tmcc_id]
        assert c.as_bytes == bytes([flags, 7])

        # info/repr should be non-empty strings
        assert isinstance(c.info, str) and c.info
        assert isinstance(repr(c), str) and repr(c)

    def test_flag_bits_combinations(self):
        # Compose flags: TAIL (11), reverse (bit2=1), train_linked (bit3), horn_masked (bit4),
        # dialog_masked (bit5), tmcc2(bit6), accessory(bit7)
        flags = 0b11111111  # all bits set
        c = ConsistComponent(tmcc_id=42, flags=flags)

        assert c.unit_type == UnitBits.TAIL
        assert c.is_tail is True
        assert c.is_reverse is True
        assert c.is_forward is False
        assert c.is_train_linked is True
        assert c.is_horn_masked is True
        assert c.is_dialog_masked is True
        assert c.is_tmcc2 is True
        assert c.is_accessory is True

    def test_from_bytes_parses_until_ff_pairs_and_reverses_order(self):
        # Build byte stream of up to 32 bytes (16 pairs), using non-FF pairs, then terminator pairs
        comps = [
            ConsistComponent(1, 0x00),
            ConsistComponent(2, 0x11),
            ConsistComponent(3, 0x22),
        ]
        # to_bytes reverses components during output; we want the raw byte order here:
        raw = b""
        for comp in reversed(comps):
            raw += comp.as_bytes
        # pad to 32 bytes with 0xFF
        raw += b"\xff" * (32 - len(raw))

        parsed = ConsistComponent.from_bytes(raw)
        # from_bytes inserts at 0, meaning reverse of encounter; given iteration i=0..,
        # it ends up reversing to original logical order comps
        assert len(parsed) == 3
        for a, b in zip(parsed, comps):
            assert a.tmcc_id == b.tmcc_id
            assert a.flags == b.flags

    def test_to_bytes_serializes_reversed_and_pads(self):
        comps = [
            ConsistComponent(10, 0x01),
            ConsistComponent(20, 0x02),
        ]
        byt = ConsistComponent.to_bytes(comps)
        assert isinstance(byt, (bytes, bytearray))
        assert len(byt) == 32

        # Expect reversed order in the leading bytes
        # Reversed: [comp2, comp1]
        expected = comps[1].as_bytes + comps[0].as_bytes
        assert byt[: len(expected)] == expected
        assert byt[len(expected) :] == b"\xff" * (32 - len(expected))


class TestRouteComponent:
    def test_from_bytes_parses_sorts_by_tmcc_id(self):
        # Create 3 components with out-of-order tmcc_ids
        r1 = RouteComponent(5, 0x00)  # switch thru
        r2 = RouteComponent(2, 0x01)  # switch out
        r3 = RouteComponent(8, 0x03)  # route
        raw = r1.as_bytes + r2.as_bytes + r3.as_bytes
        raw += b"\xff" * (32 - len(raw))

        parsed = RouteComponent.from_bytes(raw)
        # Should be sorted by tmcc_id: 2, 5, 8
        assert [r.tmcc_id for r in parsed] == [2, 5, 8]
        assert parsed[0].is_out is True
        assert parsed[1].is_thru is True
        assert parsed[2].is_route is True

    def test_to_bytes_linear_and_padded(self):
        comps = [RouteComponent(1, 0x00), RouteComponent(2, 0x01)]
        byt = RouteComponent.to_bytes(comps)
        assert isinstance(byt, (bytes, bytearray))
        assert len(byt) == 32
        expected = comps[0].as_bytes + comps[1].as_bytes
        assert byt[: len(expected)] == expected
        assert byt[len(expected) :] == b"\xff" * (32 - len(expected))

    def test_as_signature_switch_and_route(self):
        s = RouteComponent(12, 0x00)  # thru -> switch
        r = RouteComponent(7, 0x03)  # route
        assert s.as_signature == {"S12": True}
        assert r.as_signature == {"R7": True}

    def test_as_request_builds_tmcc1_switch_command(self):
        thru = RouteComponent(3, 0x00)
        out = RouteComponent(4, 0x01)

        t_req = thru.as_request
        o_req = out.as_request

        assert isinstance(t_req, CommandReq)
        assert isinstance(o_req, CommandReq)

        assert t_req.command == TMCC1SwitchCommandEnum.THRU
        assert t_req.address == 3

        assert o_req.command == TMCC1SwitchCommandEnum.OUT
        assert o_req.address == 4

    def test_as_bytes_roundtrip(self):
        r = RouteComponent(9, 0x01)
        b = r.as_bytes
        assert b == bytes([0x01, 9])



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import time
from datetime import datetime


from src.pytrain.pdi.constants import D4Action, PdiCommand
from src.pytrain.pdi.d4_req import D4Req, LIONEL_EPOCH
from src.pytrain.pdi.pdi_req import PdiReq


def test_lionel_timestamp_bytes_length_and_endianness(monkeypatch):
    # Freeze time to validate little-endian encoding
    fixed_now = LIONEL_EPOCH + 123456
    monkeypatch.setattr(time, "time", lambda: fixed_now)
    ts_bytes = D4Req.lionel_timestamp(as_bytes=True)
    assert isinstance(ts_bytes, (bytes, bytearray))
    assert len(ts_bytes) == 4
    # 123456 seconds since LIONEL_EPOCH, little endian
    assert ts_bytes == (123456).to_bytes(4, "little")


def test_build_count_request_roundtrip_engine():
    req = D4Req(0, PdiCommand.D4_ENGINE, D4Action.COUNT, count=5)
    assert req.action == D4Action.COUNT
    # COUNT requests are scoped to BASE
    from src.pytrain.protocol.constants import CommandScope

    assert req.scope == CommandScope.BASE

    packet = req.as_bytes
    parsed = PdiReq.from_bytes(packet)
    assert isinstance(parsed, D4Req)
    assert parsed.action == D4Action.COUNT
    assert parsed.count == 5
    # COUNT responses maintain BASE scope
    assert parsed.scope == CommandScope.BASE


def test_build_first_rec_request_roundtrip_train():
    req = D4Req(0, PdiCommand.D4_TRAIN, D4Action.FIRST_REC)
    packet = req.as_bytes
    parsed = PdiReq.from_bytes(packet)
    assert isinstance(parsed, D4Req)
    assert parsed.action == D4Action.FIRST_REC
    # FIRST_REC requests are sent to SYSTEM but parsed responses set BASE (see __init__ logic)
    from src.pytrain.protocol.constants import CommandScope

    assert parsed.scope == CommandScope.BASE


def test_build_map_request_roundtrip_with_tmcc_id():
    req = D4Req(0, PdiCommand.D4_ENGINE, D4Action.MAP, tmcc_id=1234)
    packet = req.as_bytes
    parsed = PdiReq.from_bytes(packet)
    assert isinstance(parsed, D4Req)
    assert parsed.action == D4Action.MAP
    # The parser extracts the TMCC id from the ASCII digits in payload
    assert parsed.tmcc_id == 1234


def test_build_query_request_with_data_bytes_and_auto_timestamp_roundtrip(monkeypatch):
    # Freeze Lionel-relative time for deterministic timestamp in payload
    fixed_now = LIONEL_EPOCH + 42
    monkeypatch.setattr(time, "time", lambda: fixed_now)

    rec_no = 42
    start = 0x10
    data_len = 3
    data_bytes = b"\x01\x02\x03"
    req = D4Req(
        rec_no,
        PdiCommand.D4_ENGINE,
        D4Action.QUERY,
        start=start,
        data_length=data_len,
        data_bytes=data_bytes,
        timestamp=None,  # force auto-stamp
    )
    packet = req.as_bytes

    parsed = PdiReq.from_bytes(packet)
    assert isinstance(parsed, D4Req)
    assert parsed.action == D4Action.QUERY
    assert parsed.record_no == rec_no
    assert parsed.start == start
    assert parsed.data_length == data_len
    # Parsed instance should keep the raw data bytes intact
    assert parsed._data_bytes == data_bytes  # pylint: disable=protected-access

    # Timestamp should be present and equal to time() - LIONEL_EPOCH
    assert parsed.timestamp == int(fixed_now - LIONEL_EPOCH)

    # Human-readable timestamp_str should reflect absolute datetime
    expected_dt = datetime.fromtimestamp(fixed_now).strftime("%Y-%m-%d %H:%M:%S")
    assert parsed.timestamp_str == expected_dt


def test_timestamp_str_empty_when_none():
    req = D4Req(1, PdiCommand.D4_ENGINE, D4Action.QUERY, start=0x00, data_length=1, timestamp=None, data_bytes=b"\x00")
    # When building, as_bytes will insert a timestamp; but the object still has timestamp=None before build()
    assert req.timestamp is None
    assert req.timestamp_str == ""


def test_as_key_structure_and_stability():
    req = D4Req(7, PdiCommand.D4_TRAIN, D4Action.COUNT, count=2)
    k1 = req.as_key
    assert isinstance(k1, tuple) and len(k1) == 4

    # Build another logically identical request and ensure key matches the same tuple components
    req2 = D4Req(7, PdiCommand.D4_TRAIN, D4Action.COUNT, count=2)
    k2 = req2.as_key
    assert k1 == k2



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

# test/utils/path_utils_tests.py
import os
from pathlib import Path

from src.pytrain.utils.path_utils import EXCLUDE, find_dir, find_file


def norm(p: str) -> str:
    # Normalize to handle platform-specific separators and any mixed use of "/"
    return os.path.normpath(p) if p is not None else None


def test_find_installation_dir():
    result = find_dir("installation")
    assert norm(result).endswith(norm(os.path.join("src", "pytrain", "installation")))


def test_find_installation_files():
    result = find_file("pytrain.bash.template")
    assert norm(result).endswith(norm(os.path.join("src", "pytrain", "installation", "pytrain.bash.template")))

    result = find_file("pytrain.service.template")
    assert norm(result).endswith(norm(os.path.join("src", "pytrain", "installation", "pytrain.service.template")))


def test_find_dir_simple(tmp_path: Path):
    # tmp_path/
    #   a/
    #     b/
    #       target/
    a = tmp_path / "a" / "b" / "target"
    a.mkdir(parents=True)

    result = find_dir("target", (str(tmp_path),))
    assert norm(result) == norm(os.path.join(str(tmp_path / "a" / "b"), "target"))


def test_find_dir_skips_hidden_and_excluded_dirs(tmp_path: Path):
    # Create hidden and excluded dirs containing the target name
    hidden = tmp_path / ".hidden" / "target"
    hidden.mkdir(parents=True)
    excluded = tmp_path / "__pycache__" / "target"
    excluded.mkdir(parents=True)
    # Create a valid visible dir with target
    visible = tmp_path / "visible" / "target"
    visible.mkdir(parents=True)

    result = find_dir("target", (str(tmp_path),))
    assert norm(result) == norm(os.path.join(str(tmp_path / "visible"), "target"))


def test_find_file_simple(tmp_path: Path):
    # tmp_path/
    #   x/
    #     y/
    #       foo.txt
    p = tmp_path / "x" / "y"
    p.mkdir(parents=True)
    fpath = p / "foo.txt"
    fpath.write_text("hello")

    result = find_file("foo.txt", (str(tmp_path),))
    assert norm(result) == norm(str(fpath))


def test_find_file_skips_hidden_files(tmp_path: Path):
    # A hidden file should not be returned even if requested
    p = tmp_path / "data"
    p.mkdir()
    hidden_file = p / ".secret"
    hidden_file.write_text("secret")

    result = find_file(".secret", (str(tmp_path),))
    assert result is None


def test_find_file_skips_excluded_dirs(tmp_path: Path):
    # Put the same filename in an excluded dir and in a valid dir;
    # the one in excluded dir should be ignored.
    excluded_dir_name = "__pycache__"
    assert excluded_dir_name in EXCLUDE

    ex_dir = tmp_path / excluded_dir_name
    ex_dir.mkdir()
    (ex_dir / "sample.bin").write_bytes(b"\x00\x01")

    ok_dir = tmp_path / "ok"
    ok_dir.mkdir()
    ok_file = ok_dir / "sample.bin"
    ok_file.write_bytes(b"\x02\x03")

    result = find_file("sample.bin", (str(tmp_path),))
    assert norm(result) == norm(str(ok_file))


def test_find_dir_not_found(tmp_path: Path):
    (tmp_path / "a").mkdir()
    result = find_dir("does-not-exist", (str(tmp_path),))
    assert result is None


def test_find_file_not_found(tmp_path: Path):
    (tmp_path / "a").mkdir()
    result = find_file("nope.txt", (str(tmp_path),))
    assert result is None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

# test/pdi/test_pdi_req_ai.py

from src.pytrain.pdi.constants import PDI_EOP, PDI_SOP, PDI_STF, PdiCommand
from src.pytrain.pdi.pdi_req import PdiReq, PingReq, TmccReq
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import CommandScope
from src.pytrain.protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum as Engine2


# noinspection PyTypeChecker
def test_encode_decode_text_roundtrip_ascii_and_nulls():
    # Normal ASCII text, padded to field length
    txt = "ABC"
    enc = PdiReq.encode_text(txt, 5)
    assert isinstance(enc, (bytes, bytearray))
    assert enc == b"ABC\x00\x00"
    dec = PdiReq.decode_text(enc)
    assert dec == "ABC"

    # None encodes to all zero bytes; decoder stops at first zero and returns empty string
    enc_none = PdiReq.encode_text(None, 4)
    assert enc_none == b"\x00\x00\x00\x00"
    dec_none = PdiReq.decode_text(enc_none)
    assert dec_none == ""

    # All 0xFF payload decodes to None
    dec_ffs = PdiReq.decode_text(b"\xff\xff\xff\xff")
    assert dec_ffs is None


def test_scope_record_length_map():
    assert PdiReq.scope_record_length(CommandScope.ENGINE) == 0xC0
    assert PdiReq.scope_record_length(CommandScope.TRAIN) == 0xC0
    assert PdiReq.scope_record_length(CommandScope.ACC) == 0xC0
    assert PdiReq.scope_record_length(CommandScope.SWITCH) == 0x40
    assert PdiReq.scope_record_length(CommandScope.ROUTE) == 0x80


def test_ping_req_as_bytes_framing_and_checksum():
    req = PingReq()
    bs = req.as_bytes

    # Basic framing
    assert isinstance(bs, (bytes, bytearray))
    assert len(bs) >= 4  # SOP + payload (>=1) + checksum + EOP
    assert bs[0] == PDI_SOP
    assert bs[-1] == PDI_EOP

    # Correct command ID
    assert bs[1] == PdiCommand.PING.value

    # Checksum validation using the receiver-side calculation (strip stuffing)
    # This mirrors how PdiReq.__init__ validates incoming frames
    payload_no_sop_eop_cs = bs[1:-2]
    _, computed_checksum = PdiReq._calculate_checksum(payload_no_sop_eop_cs, add_stf=False)
    assert bs[-2] == int.from_bytes(computed_checksum, "big")


# noinspection PyProtectedMember
def _extract_payload_from_pdi_packet(packet: bytes) -> bytes:
    """
    Helper: Strip SOP/EOP and verify checksum using the receiver path.
    Returns the de-stuffed payload (pdi_cmd + tmcc bytes).
    """
    assert packet[0] == PDI_SOP and packet[-1] == PDI_EOP
    recv = packet[1:-2]
    payload, checksum = PdiReq._calculate_checksum(recv, add_stf=False)
    assert packet[-2] == int.from_bytes(checksum, "big")
    return payload


def test_tmcc_as_packets_three_byte_tmcc2_command():
    # 3-byte TMCC2 engine command (e.g., ring bell) at 1-2 digit address
    tmcc = CommandReq.build(Engine2.RING_BELL, address=7)
    tmcc_bytes = tmcc.as_bytes
    assert len(tmcc_bytes) == 3  # standard TMCC/TMCC2 short command

    packets = TmccReq.as_packets(tmcc)
    assert isinstance(packets, list)
    assert len(packets) == 1

    pkt = packets[0]
    # Extract and validate de-stuffed payload (pdi_cmd + tmcc_bytes)
    payload = _extract_payload_from_pdi_packet(pkt)
    assert len(payload) == 1 + len(tmcc_bytes)
    assert payload[0] == PdiCommand.TMCC_TX.value
    assert payload[1:] == tmcc_bytes


def test_tmcc_as_packets_four_digit_tmcc4_command():
    # 4-digit address should produce TMCC4_TX packets with 7-byte TMCC payload (3 + 4 ASCII digits)
    tmcc_4d = CommandReq.build(Engine2.SPEED_MEDIUM, address=1234)
    tmcc_4d_bytes = tmcc_4d.as_bytes
    assert len(tmcc_4d_bytes) == 7  # 3 + 4 ASCII digits

    packets = TmccReq.as_packets(tmcc_4d)
    assert len(packets) == 1

    pkt = packets[0]
    payload = _extract_payload_from_pdi_packet(pkt)
    assert payload[0] == PdiCommand.TMCC4_TX.value
    assert payload[1:] == tmcc_4d_bytes


def test_calculate_checksum_adds_stuff_bytes_for_reserved_values():
    # Build a payload that includes reserved values to trigger stuffing:
    # payload layout: [pdi_cmd, arbitrary, SOP, arbitrary, EOP, STF]
    payload = bytes(
        [
            PdiCommand.TMCC_TX.value,
            0x12,
            PDI_SOP,  # reserved -> should be preceded by STF
            0x34,
            PDI_EOP,  # reserved -> should be preceded by STF
            PDI_STF,  # reserved -> should be preceded by STF
        ]
    )

    stuffed_stream, checksum = PdiReq._calculate_checksum(payload, add_stf=True)
    assert isinstance(stuffed_stream, (bytes, bytearray))
    assert isinstance(checksum, (bytes, bytearray))
    assert len(checksum) == 1

    # Verify that each reserved byte appears in the stuffed stream, immediately
    # preceded by a STF byte (0xFD)
    # We scan forward through stuffed_stream to match each reserved byte in order.
    sidx = 0
    for b in payload:
        if b in (PDI_SOP, PDI_EOP, PDI_STF):
            # find next occurrence of STF followed by this reserved byte
            # starting at sidx
            pos = stuffed_stream.find(bytes([PDI_STF, b]), sidx)
            assert pos != -1, f"Expected STF before reserved byte {hex(b)}"
            sidx = pos + 2
        else:
            # non-reserved bytes appear as-is; find next occurrence in order
            pos = stuffed_stream.find(bytes([b]), sidx)
            assert pos != -1, f"Expected to find byte {hex(b)} in stuffed stream"
            sidx = pos + 1



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

from __future__ import annotations

import threading

from src.pytrain.utils.singleton import singleton


@singleton
class Foo:
    _init_count: int = 0

    def __init__(self, value: int = 0) -> None:
        self.__class__._init_count += 1
        self.value = value


@singleton
class Bar:
    _init_count = 0

    def __init__(self, name: str = "") -> None:
        self.__class__._init_count += 1
        if hasattr(self, "_initialized") and getattr(self, "_initialized"):
            return
        self.name = name
        self._initialized = True


def test_returns_same_instance_and_init_called_once():
    # First construction sets value
    a = Foo(123)
    # Subsequent constructions return the same instance and do not re-run __init__
    b = Foo(456)

    assert a is b
    assert a._init_count == 1
    # Value should remain from the first initialization
    assert a.value == 123
    assert b.value == 123

    # _initialized should be set to False by the decorator
    assert hasattr(a, "_initialized")
    assert a._initialized is False


def test_thread_safety_singleton_single_instance():
    instances = []
    lock = threading.Lock()

    def build_instance(i: int):
        inst = Foo(i)
        with lock:
            instances.append(inst)

    threads = [threading.Thread(target=build_instance, args=(i,)) for i in range(50)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    # All threads should have received the exact same instance
    assert len(instances) == 50
    assert len({id(x) for x in instances}) == 1

    # __init__ must have been called only once in total
    assert instances[0]._init_count == 1
    # Value remains as set during the very first construction
    assert instances[0].value == 123


def test_different_singleton_classes_are_independent():
    f = Foo(1)
    b1 = Bar("first")
    b2 = Bar("second")

    # Bar should also be singleton
    assert b1 is b2
    assert b1._init_count == 1
    assert b1.name == "first"
    assert b2.name == "first"

    # Foo and Bar are different singletons
    assert f is not b1


def test_repeated_calls_do_not_change_state():
    f2 = Foo(42)
    f1 = Foo(999)
    # State stays as initially set
    assert f1 is f2
    assert f1.value == 123  # from the very first initialization in earlier test
    assert f1._init_count == 1


def test_initialized_flag_present_and_true_on_bar():
    a = Foo(123)
    assert hasattr(a, "_initialized")
    assert getattr(a, "_initialized") is False

    b = Bar("alpha")
    assert hasattr(b, "_initialized")
    assert getattr(b, "_initialized") is True



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import os
import select
from queue import Full

import pytest

# Import from the project package layout
from src.pytrain.utils.pollable_queue import PollableQueue


def _select_readable(fd, timeout=0.2) -> bool:
    r, _, _ = select.select([fd], [], [], timeout)
    return bool(r)


def test_fileno_is_valid() -> None:
    q = PollableQueue(4)
    fd = q.fileno()
    assert isinstance(fd, int)
    # FD should be positive and selectable
    assert fd > 0
    # Not readable initially
    assert _select_readable(fd, timeout=0.05) is False


def test_select_ready_on_put_and_clears_on_get() -> None:
    q = PollableQueue(4)
    fd = q.fileno()

    # Put an item -> should signal readability
    q.put("a")
    assert _select_readable(fd, timeout=0.5) is True

    # get() must consume the socket byte and return the item
    item = q.get()
    assert item == "a"

    # No pending signals after get
    assert _select_readable(fd, timeout=0.05) is False


def test_multiple_puts_and_gets_signal_and_drain() -> None:
    q = PollableQueue(10)
    fd = q.fileno()

    items = list(range(5))
    for x in items:
        q.put(x)

    # Readable after enqueuing
    assert _select_readable(fd, timeout=0.5) is True

    # Drain and verify order
    out = [q.get() for _ in items]
    assert out == items

    # Socket drained as well
    assert _select_readable(fd, timeout=0.05) is False


def test_put_when_full_does_not_signal_extra_bytes() -> None:
    # Fill capacity to 1
    q = PollableQueue(1)
    fd = q.fileno()

    q.put("first")
    # Should be readable for the single enqueued item
    assert _select_readable(fd, timeout=0.5) is True

    # Second put should fail and must not add a socket byte
    with pytest.raises(Full):
        q.put("second", block=False)

    # Drain the first (and only) signal and item
    assert q.get() == "first"

    # After consuming, there should be no pending readability, ensuring
    # the failed put didn't add an extra byte to the socket
    assert _select_readable(fd, timeout=0.05) is False


def test_non_posix_fallback_socket_pair(monkeypatch) -> None:
    # Force non-POSIX branch by monkeypatching os.name
    monkeypatch.setattr(os, "name", "nt", raising=False)
    q = PollableQueue(2)
    fd = q.fileno()

    # Basic signaling semantics should still hold
    assert _select_readable(fd, timeout=0.05) is False
    q.put(123)
    assert _select_readable(fd, timeout=0.5) is True
    assert q.get() == 123
    assert _select_readable(fd, timeout=0.05) is False



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import pytest

from src.pytrain.db.accessory_state import AccessoryState
from src.pytrain.pdi.amc2_req import Amc2Req
from src.pytrain.pdi.asc2_req import Asc2Req
from src.pytrain.pdi.bpc2_req import Bpc2Req
from src.pytrain.pdi.constants import Amc2Action, Asc2Action, Bpc2Action, IrdaAction, PdiCommand
from src.pytrain.pdi.irda_req import IrdaReq
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import CommandScope
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum as Aux
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1HaltCommandEnum


class TestAccessoryState:
    @staticmethod
    def _new_acc(addr: int = 9) -> AccessoryState:
        st = AccessoryState(CommandScope.ACC)
        # Allocate comp_data so as_bytes can emit state packet first
        st.initialize(CommandScope.ACC, addr)
        # Ensure address-dependent code paths are deterministic
        st._address = addr  # type: ignore[attr-defined]
        return st

    def test_invalid_scope_rejected(self):
        with pytest.raises(ValueError):
            _ = AccessoryState(CommandScope.ENGINE)

    def test_tmcc_aux_updates_and_numeric(self):
        acc = self._new_acc(12)
        # AUX1 option-one should toggle aux1 state ON and set aggregate aux_state
        req_aux1_opt1 = CommandReq.build(Aux.AUX1_OPT_ONE, acc.address)
        acc.update(req_aux1_opt1)
        assert acc.aux_state == Aux.AUX1_OPT_ONE
        assert acc.aux1_state in {Aux.AUX1_ON, Aux.AUX1_OPT_ONE, Aux.AUX1_OFF}  # implementation toggles ON/OFF
        # AUX2 ON should set aux2 explicit
        req_aux2_on = CommandReq.build(Aux.AUX2_ON, acc.address)
        acc.update(req_aux2_on)
        assert acc.aux2_state == Aux.AUX2_ON
        # NUMERIC sets number/value
        req_num = CommandReq.build(Aux.NUMERIC, acc.address, data=7)
        acc.update(req_num)
        assert acc.value == 7
        # is_known reflects presence of aux/number
        assert acc.is_known is True

    def test_bpc2_control3_sets_block_power_and_on_off(self):
        acc = self._new_acc(15)
        # First PDI packet establishes LCS source and power-district behavior
        pdi_on = Bpc2Req(acc.address, PdiCommand.BPC2_SET, Bpc2Action.CONTROL3, state=1)
        acc.update(pdi_on)
        assert acc.is_lcs_component is True
        assert acc.is_power_district is True
        assert acc.is_asc2 is False
        assert acc.is_amc2 is False
        assert acc.aux_state == Aux.AUX1_OPT_ONE
        assert acc.aux1_state == Aux.AUX1_ON and acc.aux2_state == Aux.AUX2_ON

        # Turn OFF
        pdi_off = Bpc2Req(acc.address, PdiCommand.BPC2_SET, Bpc2Action.CONTROL3, state=0)
        acc.update(pdi_off)
        assert acc.aux_state == Aux.AUX2_OPT_ONE
        assert acc.aux1_state == Aux.AUX1_OFF and acc.aux2_state == Aux.AUX2_OFF

        # as_dict for power district
        d = acc.as_dict()
        assert d["type"] == "power district"
        assert d["block"] in {"on", "off"}

    def test_bpc2_control1_causes_error(self):
        acc = self._new_acc(15)
        # PDI packet should be rejected, as scope is TRAIN
        pdi_on = Bpc2Req(acc.address, PdiCommand.BPC2_SET, Bpc2Action.CONTROL1, state=1)
        assert pdi_on.scope == CommandScope.TRAIN
        with pytest.raises(AttributeError):
            acc.update(pdi_on)

    def test_asc2_control1_marks_lcs_on_power(self):
        acc = self._new_acc(6)
        asc2_on = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1)
        acc.update(asc2_on)
        assert acc.is_lcs_component is True
        assert acc.is_asc2 is True
        assert acc.is_power_district is False
        assert acc.aux_state is acc.aux_state == Aux.AUX1_OPT_ONE
        # off
        asc2_off = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=0)
        acc.update(asc2_off)
        assert acc.aux_state == Aux.AUX2_OPT_ONE

    def test_amc2_motor_marks_lcs_on_power(self):
        config = Amc2Req.from_bytes(bytes.fromhex("d1461903190000000202020303010100000000000000000077df"))
        acc = self._new_acc(25)
        acc.update(config)
        assert acc.is_lcs_component is True
        assert acc.is_amc2 is True
        assert acc.is_known
        assert acc.aux_state == Aux.AUX2_OPT_ONE
        assert acc.aux1_state == Aux.AUX1_OFF
        assert acc.aux2_state == Aux.AUX2_OFF

        assert acc.motor1
        assert acc.motor1.speed == 0
        assert acc.motor1.state is False
        assert acc.motor1.restore_state is False

        assert acc.motor2
        assert acc.motor2.speed == 0
        assert acc.motor2.state is False
        assert acc.motor2.restore_state is False

        amc2_sp = Amc2Req(acc.address, PdiCommand.AMC2_SET, Amc2Action.MOTOR, motor=1, speed=100)
        acc.update(amc2_sp)
        acc.motor2.restore_state = True

        assert acc.motor2
        assert acc.motor2.speed == 100
        assert acc.motor2.state
        assert acc.motor2.restore_state is True

        assert acc.motor1
        assert acc.motor1.speed == 0
        assert acc.motor1.state is False
        assert acc.motor1.restore_state is False

        amc2_sp = Amc2Req(acc.address, PdiCommand.AMC2_SET, Amc2Action.MOTOR, motor=0, speed=50)
        acc.update(amc2_sp)
        acc.motor1.restore_state = True

        assert acc.motor1.speed == 50
        assert acc.motor2.speed == 100
        assert acc.motor1.state
        assert acc.motor2.state

    def test_irda_marks_sensor_track_and_serialization_includes_irda(self):
        acc = self._new_acc(18)
        # Mark as sensor track via IRDA INFO RX
        ir = IrdaReq(acc.address, PdiCommand.IRDA_RX, IrdaAction.INFO, scope=CommandScope.ACC)
        acc.update(ir)
        assert acc.is_sensor_track is True

        # as_bytes: first emits comp_data state packet; then an IRDA state req packet
        blob = acc.as_bytes()
        assert isinstance(blob, (bytes, bytearray))
        # Should contain IRDA header byte (0x32) somewhere after the first packet
        assert PdiCommand.IRDA_RX.as_bytes in blob

    def test_as_bytes_tmcc_accessory_commands_append(self):
        acc = self._new_acc(22)
        # Issue some TMCC accessory commands (not LCS)
        acc.update(CommandReq.build(Aux.AUX1_ON, acc.address))
        acc.update(CommandReq.build(Aux.AUX2_OFF, acc.address))

        bs = acc.as_bytes()
        assert isinstance(bs, (bytes, bytearray))
        # It should include TMCC3-byte packets for aux1/aux2 after the comp_data pdi block
        # Check presence by building expected command bytes
        aux1_bytes = CommandReq.build(Aux.AUX1_ON, acc.address).as_bytes
        aux2_bytes = CommandReq.build(Aux.AUX2_OFF, acc.address).as_bytes
        assert aux1_bytes in bs
        assert aux2_bytes in bs

    def test_as_dict_accessory_and_sensor_types(self):
        # Accessory default
        acc = self._new_acc(5)
        acc.update(CommandReq.build(Aux.AUX2_OPT_ONE, acc.address))
        d1 = acc.as_dict()
        assert d1["type"] == "accessory"
        # Sensor track
        acc2 = self._new_acc(7)
        acc2.update(IrdaReq(acc2.address, PdiCommand.IRDA_RX, IrdaAction.INFO, scope=CommandScope.ACC))
        d2 = acc2.as_dict()
        assert d2["type"] == "sensor track"

    def test_payload_renders_reasonable_info(self):
        acc = self._new_acc(30)
        # Non-LCS accessory, AUX1 toggled and number set
        acc.update(CommandReq.build(Aux.AUX1_OPT_ONE, acc.address))
        acc.update(CommandReq.build(Aux.NUMERIC, acc.address, data=3))
        s = acc.payload
        assert isinstance(s, str)
        assert "Aux" in s or "Unknown" in s
        assert "Aux Num:" in s

    def test_halt_resets_aux_and_number(self):
        acc = self._new_acc(31)
        # Turn on via TMCC and set a number
        acc.update(CommandReq.build(Aux.AUX1_OPT_ONE, acc.address))
        acc.update(CommandReq.build(Aux.NUMERIC, acc.address, data=4))
        assert acc.aux_state == Aux.AUX1_OPT_ONE
        assert acc.value == 4

        # HALT should force OFF and clear number
        halt = CommandReq.build(TMCC1HaltCommandEnum.HALT, acc.address)
        acc.update(halt)
        assert acc.aux_state == Aux.AUX2_OPT_ONE
        assert acc.aux1_state == Aux.AUX1_OFF
        assert acc.aux2_state == Aux.AUX2_OFF
        assert acc.value is None

    def test_tmcc_commands_ignored_after_pdi_source(self):
        acc = self._new_acc(32)
        # Establish LCS source via ASC2, ON
        asc2_on = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1)
        acc.update(asc2_on)
        assert acc.is_lcs_component is True
        assert acc.aux_state == Aux.AUX1_OPT_ONE
        assert acc.aux1_state == Aux.AUX1_ON
        assert acc.aux2_state == Aux.AUX2_ON

        # TMCC accessory commands should NOT change state once LCS source
        acc.update(CommandReq.build(Aux.AUX1_OFF, acc.address))
        acc.update(CommandReq.build(Aux.AUX2_OFF, acc.address))
        # Still the LCS-determined ON state
        assert acc.aux_state == Aux.AUX1_OPT_ONE
        assert acc.aux1_state == Aux.AUX1_ON
        assert acc.aux2_state == Aux.AUX2_ON

    def test_payload_for_asc2_includes_prefix_for_on_and_off(self):
        # ON
        acc_on = self._new_acc(33)
        acc_on.update(Asc2Req(acc_on.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1))
        s_on = acc_on.payload
        assert isinstance(s_on, str)
        # Should include "Asc2 " and reflect ON
        assert "Asc2" in s_on

        # OFF
        acc_off = self._new_acc(34)
        acc_off.update(Asc2Req(acc_off.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=0))
        s_off = acc_off.payload
        assert isinstance(s_off, str)
        # Should include "Asc2 " and reflect OFF, not just bare "OFF"
        assert "Asc2" in s_off

    def test_as_bytes_lcs_asc2_uses_first_action_and_value(self):
        acc = self._new_acc(35)
        # First PDI action (CONTROL1) ON should be serialized accordingly
        asc2_on = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1)
        acc.update(asc2_on)
        bs_on = acc.as_bytes()
        # Expected serialized action should be present
        expected_on = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1).as_bytes
        assert expected_on in bs_on

        # Now OFF updates the aux state; as_bytes should reflect OFF
        asc2_off = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=0)
        acc.update(asc2_off)
        bs_off = acc.as_bytes()
        expected_off = Asc2Req(acc.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=0).as_bytes
        assert expected_off in bs_off

    def test_as_bytes_amc2_includes_numeric_when_set(self):
        acc = self._new_acc(25)
        # Provide AMC2 config to establish AMC2/LCS context
        cfg = Amc2Req.from_bytes(bytes.fromhex("d1461903190000000202020303010100000000000000000077df"))
        acc.update(cfg)
        assert acc.is_amc2 is True and acc.is_lcs_component is True

        # Set a number; for AMC2, as_bytes should include the numeric packet
        acc.update(CommandReq.build(Aux.NUMERIC, acc.address, data=5))
        blob = acc.as_bytes()
        numeric_bytes = CommandReq(Aux.NUMERIC, acc.address, 5).as_bytes
        assert numeric_bytes in blob

    def test_get_motor_and_lamps_none_when_not_amc2(self):
        acc = self._new_acc(37)
        # Plain accessory
        assert acc.get_motor(1) is None
        assert acc.get_motor(2) is None
        assert acc.get_lamp(1) is None
        assert acc.get_lamp(2) is None
        assert acc.get_lamp(3) is None
        assert acc.get_lamp(4) is None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

#
#  Tests for src/pytrain/pdi/constants.py
#
import itertools

import pytest

from src.pytrain.pdi.constants import (
    ALL_GETs,
    ALL_RXs,
    ALL_SETs,
    Amc2Action,
    Asc2Action,
    Bpc2Action,
    CommonAction,
    D4Action,
    PdiCommand,
    Ser2Action,
    Stm2Action,
    WiFiAction,
)


def test_pdi_command_basic_classification_flags():
    # Spot-check a variety of commands and their flags
    assert PdiCommand.PING.is_ping is True
    assert PdiCommand.PING.is_tmcc is False
    assert PdiCommand.TMCC_TX.is_tmcc is True
    assert PdiCommand.TMCC4_RX.is_tmcc is True

    # Base family classified as base/sendable
    for cmd in (
        PdiCommand.BASE_ENGINE,
        PdiCommand.BASE_TRAIN,
        PdiCommand.BASE_ACC,
        PdiCommand.BASE,
        PdiCommand.BASE_ROUTE,
        PdiCommand.BASE_SWITCH,
        PdiCommand.BASE_MEMORY,
        PdiCommand.UPDATE_ENGINE_SPEED,
        PdiCommand.UPDATE_TRAIN_SPEED,
    ):
        assert cmd.is_base is True
        assert cmd.is_sendable is True

    # D4 family
    assert PdiCommand.D4_ENGINE.is_d4 is True
    assert PdiCommand.D4_TRAIN.is_d4 is True
    assert PdiCommand.D4_ENGINE.is_receivable is True  # d4 are receivable
    assert PdiCommand.D4_TRAIN.is_sendable is True  # and sendable

    # GET/SET/RX suffix helpers
    assert PdiCommand.WIFI_GET.is_get is True
    assert PdiCommand.WIFI_SET.is_set is True
    assert PdiCommand.WIFI_RX.is_receive is True

    # Domain helpers
    assert PdiCommand.IRDA_GET.is_irda is True
    assert PdiCommand.WIFI_PING.is_wifi is True
    assert PdiCommand.ASC2_RX.is_asc2 is True
    assert PdiCommand.AMC2_SET.is_amc2 is True
    assert PdiCommand.STM2_GET.is_stm2 is True
    assert PdiCommand.SER2_RX.is_ser2 is True
    assert PdiCommand.BPC2_SET.is_bpc2 is True
    assert PdiCommand.BLOCK_RX.is_block is True

    # LCS grouping (any of these implies LCS)
    for cmd in (
        PdiCommand.WIFI_GET,
        PdiCommand.ASC2_SET,
        PdiCommand.IRDA_RX,
        PdiCommand.SER2_GET,
        PdiCommand.BPC2_RX,
        PdiCommand.STM2_SET,
        PdiCommand.AMC2_RX,
    ):
        assert cmd.is_lcs is True

    # as_bytes returns single byte
    assert PdiCommand.PING.as_bytes == bytes([PdiCommand.PING.value])


def test_pdi_command_collections_suffix_sets_are_consistent():
    # The suffix-derived sets should match name endings for the enum
    rx_by_name = {e for e in PdiCommand if e.name.endswith("_RX")}
    set_by_name = {e for e in PdiCommand if e.name.endswith("_SET")}
    get_by_name = {e for e in PdiCommand if e.name.endswith("_GET")}
    assert ALL_RXs == rx_by_name
    assert ALL_SETs == set_by_name
    assert ALL_GETs == get_by_name


def test_pdi_action_common_properties_and_bytes():
    # Every PdiAction should expose .bits and .as_bytes consistent with ActionDef
    all_action_enums = [CommonAction, WiFiAction, Asc2Action, Amc2Action, Ser2Action, Bpc2Action, Stm2Action, D4Action]
    for enum_cls in all_action_enums:
        for action in enum_cls:
            # bits must be int and in byte range 0..255 for to_bytes(1)
            assert isinstance(action.bits, int)
            assert 0 <= action.bits <= 0xFF
            assert action.as_bytes == action.bits.to_bytes(1, "big")
            # __repr__ contains the friendly title and the flags summary
            r = repr(action)
            assert action.title in r
            # flags are rendered as g/x, s/x, r/x
            assert any(flag in r for flag in ("g", "x", "s"))
            if action.is_gettable:
                assert action.opts.count("g") == 1
            if action.is_settable:
                assert action.opts.count("s") == 1
            if action.is_responses:
                assert action.opts.count("r") == 1


def test_common_action_is_config_detection():
    # is_config is specialized: compares bits to CommonAction.CONFIG.bits
    for act in CommonAction:
        if act is CommonAction.CONFIG:
            assert act.is_config is True
        else:
            assert act.is_config is False


def test_d4_action_capabilities_matrix():
    # D4Action definitions specify gettable/settable/responds; verify a few
    expected = {
        D4Action.QUERY: (True, False, True),
        D4Action.UPDATE: (False, True, True),
        D4Action.NEXT_REC: (True, False, True),
        D4Action.CLEAR: (False, True, True),
        D4Action.MAP: (True, False, True),
        D4Action.FIRST_REC: (True, False, True),
        D4Action.COUNT: (True, False, True),
    }
    for act, (g, s, r) in expected.items():
        val = act.value  # ActionDef
        assert val.is_gettable is g
        assert val.is_settable is s
        assert val.is_responses is r


@pytest.mark.parametrize(
    "enum_cls",
    [CommonAction, WiFiAction, Asc2Action, Amc2Action, Ser2Action, Bpc2Action, Stm2Action, D4Action],
)
def test_pdi_action_repr_has_option_flags(enum_cls):
    # Ensure repr of actions renders the three flags indicators
    for action in enum_cls:
        rep = repr(action)
        # ensure the three-character flags like [gsr]/[xxr]/etc appear
        assert "[" in rep and "]" in rep
        # at least 3 of g/x/s/x/r/x markers total
        flag_chunk = rep.split("[")[-1].split("]")[0]
        assert len(flag_chunk) == 3
        assert set(flag_chunk).issubset({"g", "s", "r", "x"})


def test_action_bytes_do_not_collide_within_each_enum():
    # Within a given action enum class, bits should be unique to avoid ambiguity
    for enum_cls in (CommonAction, WiFiAction, Asc2Action, Amc2Action, Ser2Action, Bpc2Action, Stm2Action, D4Action):
        bits = [a.bits for a in enum_cls]
        assert len(bits) == len(set(bits)), f"Duplicate action bits in {enum_cls.__name__}"


def test_pdi_command_sendable_receivable_logic_spotchecks():
    # Sendable: base, d4, *_GET, *_SET
    # Receivable: *_RX or d4
    assert PdiCommand.WIFI_GET.is_sendable is True
    assert PdiCommand.WIFI_SET.is_sendable is True
    assert PdiCommand.WIFI_RX.is_sendable is False
    assert PdiCommand.WIFI_RX.is_receivable is True

    assert PdiCommand.D4_ENGINE.is_sendable is True
    assert PdiCommand.D4_ENGINE.is_receivable is True

    # Non-lcs standalones
    assert PdiCommand.PING.is_sendable is False
    assert PdiCommand.PING.is_receivable is False


def test_as_bytes_single_byte_roundtrip_subset():
    # Basic sanity: the bytes value matches integer value of the command
    for cmd in itertools.islice(iter(PdiCommand), 0, 10):
        assert cmd.as_bytes == bytes([cmd.value])



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import threading

import pytest

from src.pytrain.db.client_state_listener import ClientStateHandler, ClientStateListener
from src.pytrain.pdi.constants import PDI_EOP, PDI_SOP, PDI_STF


@pytest.fixture(autouse=True)
def reset_singleton_state():
    # Ensure singleton state doesn't leak between tests
    ClientStateListener._instance = None
    yield
    ClientStateListener._instance = None


class FakeCommBufferProxy:
    def __init__(self, version=None):
        self._server_version = version
        self._ev = threading.Event()
        self._ev.set()
        self._registered = []
        self._synced = False
        self._hb = False

    @staticmethod
    def server_port():
        return 12345

    def register(self, port):
        self._registered.append(port)

    def sync_state(self):
        self._synced = True

    def start_heart_beat(self):
        self._hb = True

    def server_version_available(self):
        return self._ev

    @property
    def server_version(self):
        return self._server_version

    @server_version.setter
    def server_version(self, v):
        self._server_version = v


class FakeListener:
    def __init__(self):
        self.offered = []
        self.subscribed = []
        self.unsubscribed = []

    def offer(self, data):
        self.offered.append(data)

    def subscribe(self, listener, channel, address=None, command=None, data=None):
        self.subscribed.append((listener, channel, address, command, data))

    def unsubscribe(self, listener, channel, address=None, command=None, data=None):
        self.unsubscribed.append((listener, channel, address, command, data))


def build_minimal_csl_with_fakes():
    """
    Build a ClientStateListener instance without running __init__, then inject the fakes
    we need to exercise utility methods (subscribe/unsubscribe/offer).
    """
    inst = ClientStateListener.__new__(ClientStateListener)
    inst._initialized = True
    inst._tmcc_listener = FakeListener()
    inst._pdi_listener = FakeListener()
    inst._tmcc_buffer = FakeCommBufferProxy()
    inst._port = 10000
    inst._is_running = False
    return inst


def test_singleton_build_returns_same_instance(monkeypatch):
    # Avoid running real __init__ internals by mocking __init__ to be a no-op for this test
    def _noop_init(self):
        if getattr(self, "_initialized", False):
            return
        self._initialized = True

    monkeypatch.setattr(ClientStateListener, "__init__", _noop_init)

    a = ClientStateListener.build()
    b = ClientStateListener.build()
    assert a is b, "ClientStateListener.build should return singleton instance"


def test_offer_routes_to_pdi_when_sop(monkeypatch):
    csl = build_minimal_csl_with_fakes()

    data_pdi = bytes([PDI_SOP, 0x10, 0x20, PDI_EOP])
    data_tmcc = b"\xf0\x12\x34"

    # Route PDI
    csl.offer(data_pdi)
    assert csl._pdi_listener.offered[-1] == data_pdi
    # Route TMCC
    csl.offer(data_tmcc)
    assert csl._tmcc_listener.offered[-1] == data_tmcc


# noinspection PyTypeChecker
def test_subscribe_and_unsubscribe_delegate_to_both_listeners():
    csl = build_minimal_csl_with_fakes()
    fake_sub = object()
    channel = object()

    csl.subscribe(fake_sub, channel, address=1, command=None, data=None)
    assert csl._tmcc_listener.subscribed[-1][0] is fake_sub
    assert csl._pdi_listener.subscribed[-1][0] is fake_sub

    csl.unsubscribe(fake_sub, channel, address=1, command=None, data=None)
    assert csl._tmcc_listener.unsubscribed[-1][0] is fake_sub
    assert csl._pdi_listener.unsubscribed[-1][0] is fake_sub


def test_update_client_if_needed_compares_versions(monkeypatch):
    # Make an instance without running __init__
    csl = build_minimal_csl_with_fakes()

    # Monkeypatch pytrain.get_version and get_version_tuple used inside the method
    import src.pytrain as pytrain_pkg

    monkeypatch.setattr(pytrain_pkg, "get_version", lambda: "1.2.3")
    monkeypatch.setattr(pytrain_pkg, "get_version_tuple", lambda: (1, 2, 3))

    # Case 1: server newer -> True
    csl._tmcc_buffer.server_version = (1, 3, 0)
    assert csl.update_client_if_needed(do_upgrade=False) is True

    # Case 2: server equal -> False
    csl._tmcc_buffer.server_version = (1, 2, 3)
    assert csl.update_client_if_needed(do_upgrade=False) is False

    # Case 3: server None -> True
    csl._tmcc_buffer.server_version = None
    assert csl.update_client_if_needed(do_upgrade=False) is True


# noinspection PyTypeChecker, PyUnusedLocal
def test_client_state_handler_splits_pdi_then_tmcc(monkeypatch):
    # Build a fake CSL that only records offers
    class FakeCSL:
        def __init__(self):
            self.offers = []

        def offer(self, data):
            self.offers.append(bytes(data))

    fake_csl = FakeCSL()

    # Patch ClientStateListener.build() to return our fake without touching threads/init
    monkeypatch.setattr(ClientStateListener, "build", classmethod(lambda cls: fake_csl))

    # Create a PDI frame with stuffed EOP sequence inside; then TMCC bytes
    # PDI: [SOP, 0xAA, STF, EOP, 0xBB, EOP]  -> last EOP is actual terminator
    pdi_payload = bytes([PDI_SOP, 0xAA, PDI_STF, PDI_EOP, 0xBB, PDI_EOP])
    tmcc_payload = b"\xf0\x12\x34"
    combined = pdi_payload + tmcc_payload

    # Fake socket request that returns data then empty
    # noinspection PyUnusedLocal
    class FakeReq:
        def __init__(self, chunks):
            self._chunks = list(chunks)
            self.sent = []

        def recv(self, size):
            return self._chunks.pop(0) if self._chunks else b""

        def sendall(self, data):
            self.sent.append(data)

    # Build a handler instance with our fake request; BaseRequestHandler would call handle
    handler = ClientStateHandler.__new__(ClientStateHandler)
    handler.request = FakeReq([combined])
    handler.client_address = ("127.0.0.1", 9999)
    handler.server = object()

    # Invoke handle directly
    ClientStateHandler.handle(handler)

    # It should ack once for the single recv
    assert handler.request.sent == [b"ack"]

    # Two offers: first the PDI segment, then the TMCC bytes
    assert len(fake_csl.offers) == 2
    assert fake_csl.offers[0] == pdi_payload  # exact PDI frame
    assert fake_csl.offers[1] == tmcc_payload


def test_handler_handles_multiple_recv_chunks(monkeypatch):
    # Verify that handler concatenates chunks before splitting
    class FakeCSL:
        def __init__(self):
            self.offers = []

        def offer(self, data):
            self.offers.append(bytes(data))

    fake_csl = FakeCSL()
    monkeypatch.setattr(ClientStateListener, "build", classmethod(lambda cls: fake_csl))

    pdi = bytes([PDI_SOP, 0x01, 0x02, PDI_EOP])
    tmcc = b"\xf0\xab\xcd"
    chunks = [pdi[:2], pdi[2:], tmcc, b""]  # fragmented receive  # noqa: F841

    # noinspection PyUnusedLocal,PyShadowingNames
    class FakeReq:
        def __init__(self, chunks):
            self._chunks = list(chunks)
            self.sent = []

        def recv(self, size):
            return self._chunks.pop(0) if self._chunks else b""

        def sendall(self, data):
            self.sent.append(data)

    handler = ClientStateHandler.__new__(ClientStateHandler)
    handler.request = FakeReq



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
from __future__ import annotations

import atexit
import logging
from abc import ABC, ABCMeta, abstractmethod
from queue import Queue, Empty
from threading import Condition, Event, RLock, Thread, get_ident
from tkinter import TclError
from typing import Callable, Generic, TypeVar, cast, Any

from guizero import App, Box, Combo, PushButton, Text, Slider
from guizero.base import Widget
from guizero.event import EventData

from ..comm.command_listener import CommandDispatcher
from ..db.accessory_state import AccessoryState
from ..db.component_state import ComponentState, RouteState, SwitchState
from ..db.component_state_store import ComponentStateStore
from ..db.state_watcher import StateWatcher
from ..gpio.gpio_handler import GpioHandler
from ..pdi.amc2_req import Amc2Req
from ..pdi.asc2_req import Asc2Req
from ..pdi.constants import Asc2Action, PdiCommand, Amc2Action
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum, TMCC1RouteCommandEnum, TMCC1SwitchCommandEnum
from ..utils.path_utils import find_file

log = logging.getLogger(__name__)
S = TypeVar("S", bound=ComponentState)


class StateBasedGui(Thread, Generic[S], ABC):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(
        self,
        title: str,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        enabled_bg: str = "green",
        disabled_bg: str = "black",
        enabled_text: str = "black",
        disabled_text: str = "lightgrey",
        scale_by: float = 1.0,
    ) -> None:
        Thread.__init__(self, daemon=True, name=f"{title} GUI")
        self._cv = Condition(RLock())
        self._ev = Event()
        if width is None or height is None:
            try:
                from tkinter import Tk

                root = Tk()
                self.width = root.winfo_screenwidth()
                self.height = root.winfo_screenheight()
                root.destroy()
            except Exception as e:
                log.exception("Error determining window size", exc_info=e)
        else:
            self.width = width
            self.height = height
        self.title = title
        self.label = label
        self._aggrigator = aggrigator
        self._scale_by = scale_by
        self._text_size: int = 24
        self._button_pad_x = 20
        self._button_pad_y = 10
        self._button_text_pad_x = 10
        self._button_text_pad_y = 12

        self._enabled_bg = enabled_bg
        self._disabled_bg = disabled_bg
        self._enabled_text = enabled_text
        self._disabled_text = disabled_text
        self.left_arrow = find_file("left_arrow.jpg")
        self.right_arrow = find_file("right_arrow.jpg")
        self.app = self.by_name = self.by_number = self.box = self.btn_box = self.y_offset = None
        self.pd_button_height = self.pd_button_width = self.left_scroll_btn = self.right_scroll_btn = None
        self.aggrigator_combo = None
        self._max_name_len = 0
        self._max_button_rows = self._max_button_cols = None
        self._first_button_col = 0
        self.sort_func = None
        self._app_counter = 0
        self._message_queue = Queue()

        # States
        self._states = dict[int, S]()
        self._state_buttons = dict[int, PushButton]()
        self._state_watchers = dict[int, StateWatcher]()

        # Thread-aware shutdown signaling
        self._tk_thread_id: int | None = None
        self._is_closed = False
        self._shutdown_flag = Event()

        # listen for state changes
        self._dispatcher = CommandDispatcher.get()
        self._state_store = ComponentStateStore.get()
        self._synchronized = False
        self._sync_state = self._state_store.get_state(CommandScope.SYNC, 99)
        if self._sync_state and self._sync_state.is_synchronized is True:
            self._sync_watcher = None
            self.on_sync()
        else:
            self._sync_watcher = StateWatcher(self._sync_state, self.on_sync)

        # Important: don't call tkinter from atexit; only signal
        atexit.register(lambda: self._shutdown_flag.set())

    def close(self) -> None:
        # Only signal shutdown here; actual tkinter destroy happens on the GUI thread
        if not self._is_closed:
            self._is_closed = True
            self._shutdown_flag.set()

    def reset(self) -> None:
        self.close()

    @property
    def destroy_complete(self) -> Event:
        return self._ev

    # noinspection PyTypeChecker
    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            self._synchronized = True

            # get all target states; watch for state changes
            accs = self.get_target_states()
            for acc in accs:
                nl = len(acc.road_name)
                self._max_name_len = nl if nl > self._max_name_len else self._max_name_len
                self._states[acc.tmcc_id] = acc
                self._state_watchers[acc.tmcc_id] = StateWatcher(acc, self.on_state_change_action(acc.tmcc_id))

            # start GUI
            self.start()

    # noinspection PyTypeChecker
    def update_button(self, tmcc_id: int) -> None:
        with self._cv:
            pd: S = self._states[tmcc_id]
            pb = self._state_buttons[tmcc_id]
            if self.is_active(pd):
                self._set_button_active(pb)
            else:
                self._set_button_inactive(pb)

    # noinspection PyTypeChecker
    def _set_button_inactive(self, widget: Widget):
        widget.bg = self._disabled_bg
        widget.text_color = self._disabled_text

    # noinspection PyTypeChecker
    def _set_button_active(self, widget: Widget):
        widget.bg = self._enabled_bg
        widget.text_color = self._enabled_text

    def on_state_change_action(self, tmcc_id: int) -> Callable:
        def upd():
            if not self._shutdown_flag.is_set():
                self._message_queue.put((self.update_button, [tmcc_id]))

        return upd

    # noinspection PyTypeChecker
    def run(self) -> None:
        self._shutdown_flag.clear()
        self._ev.clear()
        self._tk_thread_id = get_ident()
        GpioHandler.cache_handler(self)
        self.app = app = App(title=self.title, width=self.width, height=self.height)
        app.full_screen = True
        app.when_closed = self.close

        # poll for shutdown requests from other threads; this runs on the GuiZero/Tk thread
        def _poll_shutdown():
            self._app_counter += 1
            if self._shutdown_flag.is_set():
                try:
                    app.destroy()
                except TclError:
                    pass  # ignore, we're shutting down
                return None
            else:
                try:
                    message = self._message_queue.get_nowait()
                    if isinstance(message, tuple):
                        message[0](*message[1])
                except Empty:
                    pass
            return None

        app.repeat(25, _poll_shutdown)

        self.box = box = Box(app, layout="grid")
        app.bg = box.bg = "white"

        ts = self._text_size
        _ = Text(box, text=" ", grid=[0, 0, 6, 1], size=6, height=1, bold=True)
        _ = Text(box, text="    ", grid=[1, 1], size=ts)
        if self._aggrigator:
            txt_lbl = txt_spacer = None
            ats = int(round(23 * self._scale_by))
            ag_box = Box(box, grid=[2, 1, 2, 1], layout="auto")
            if self.label:
                # Wrap the Text in a vertical container so we can insert a spacer above it
                txt_vbox = Box(ag_box, layout="auto", align="left")
                txt_spacer = Box(txt_vbox, height=1, width=1)  # will be set after measuring
                txt_lbl = Text(txt_vbox, text=self.label + ": ", align="top", size=ats, bold=True)
            # Wrap the Combo in a vertical container as well
            combo_vbox = Box(ag_box, layout="auto", align="right")
            combo_spacer = Box(combo_vbox)  # will be set after measuring
            self.aggrigator_combo = Combo(
                combo_vbox,
                options=self._aggrigator.guis,
                selected=self.title,
                align="top",
                command=self.on_combo_change,
            )
            self.aggrigator_combo.text_size = ats
            self.aggrigator_combo.text_bold = True

            # Compute center-offset and add top spacer to the shorter control
            if self.label:
                # After creation, force layout and measure actual heights
                self.app.update()
                txt_h = txt_lbl.tk.winfo_height() if self.label else 0
                combo_h = self.aggrigator_combo.tk.winfo_height()
                if txt_h < combo_h:
                    delta = (combo_h - txt_h) // 2
                    txt_spacer.height = max(0, delta)
                    combo_spacer.height = 0
                elif combo_h < txt_h:
                    delta = (txt_h - combo_h) // 2
                    combo_spacer.height = max(0, delta)
                    txt_spacer.height = 0
                else:
                    txt_spacer.height = combo_spacer.height = 0
            else:
                # No label Text; ensure combo is not offset
                combo_spacer.height = 0
        else:
            # customize label
            label = f"{self.label} {self.title}" if self.label else self.title
            _ = Text(box, text=label, grid=[2, 1, 2, 1], size=ts, bold=True)
        _ = Text(box, text="    ", grid=[4, 1], size=ts)
        self.by_number = PushButton(
            box,
            text="By TMCC ID",
            grid=[2, 2],
            command=self.sort_by_number,
            padx=self._button_text_pad_x,
            pady=self._button_text_pad_y,
        )

        self.by_name = PushButton(
            box,
            text="By Name",
            grid=[3, 2],
            width=len("By TMCC ID"),
            command=self.sort_by_name,
            padx=self._button_text_pad_x,
            pady=self._button_text_pad_y,
        )
        self.by_name.text_size = self.by_number.text_size = int(round(18 * self._scale_by))
        self.by_number.text_bold = True

        _ = Text(box, text=" ", grid=[0, 3, 6, 1], size=4, height=1, bold=True)
        self.app.update()

        parent = self.by_number.tk.master
        parent.grid_columnconfigure(self.by_number.tk.grid_info()["column"], pad=40)
        parent.grid_columnconfigure(self.by_name.tk.grid_info()["column"], pad=40)
        parent.grid_rowconfigure(self.by_number.tk.grid_info()["row"], pad=10)

        # add scroll btns
        sort_btn_height = self.by_number.tk.winfo_height()
        self.left_scroll_btn = PushButton(
            box,
            grid=[0, 1, 1, 2],
            enabled=False,
            image=self.left_arrow,
            height=sort_btn_height * 2,
            width=sort_btn_height * 2,
            align="left",
            command=self.scroll_left,
        )
        self.right_scroll_btn = PushButton(
            box,
            grid=[5, 1, 1, 2],
            enabled=False,
            image=self.right_arrow,
            height=sort_btn_height * 2,
            width=sort_btn_height * 2,
            align="right",
            command=self.scroll_right,
        )

        self.app.update()
        self.y_offset = self.box.tk.winfo_y() + self.box.tk.winfo_height()

        # put the buttons in a separate box
        self.btn_box = Box(app, layout="grid")

        # Order by tmcc_id
        self.sort_by_number()

        # Display GUI and start event loop; call blocks
        try:
            app.display()
        except TclError:
            # If Tcl is already tearing down, ignore
            pass
        finally:
            # Explicitly drop references to tkinter/guizero objects on the Tk thread
            if self._aggrigator:
                for sw in self._state_watchers.values():
                    sw.shutdown()
                self._state_watchers.clear()
            self.aggrigator_combo = None
            self.left_scroll_btn = None
            self.right_scroll_btn = None
            self.by_name = None
            self.by_number = None
            self.btn_box = None
            self.box = None
            self._state_buttons.clear()
            self._state_buttons = None
            self.app = None
            self._ev.set()

    def on_combo_change(self, option: str) -> None:
        if option == self.title:
            return  # Noop
        else:
            self._aggrigator.cycle_gui(option)

    # noinspection PyUnusedLocal
    def _reset_state_buttons(self) -> None:
        for pdb in self._state_buttons.values():
            if not isinstance(pdb, list):
                pdb = [pdb]
            for widget in pdb:
                if hasattr(widget, "component_state"):
                    widget.component_state = None
                if hasattr(widget, "when_left_button_pressed"):
                    widget.when_left_button_pressed = None
                if hasattr(widget, "when_left_button_released"):
                    widget.when_left_button_released = None
                widget.hide()
                widget.destroy()
        self._state_buttons.clear()

    # noinspection PyTypeChecker
    def _make_state_buttons(self, states: list[S] = None) -> None:
        with self._cv:
            if self._state_buttons:
                self._reset_state_buttons()
            active_cols = {self._first_button_col, self._first_button_col + 1}
            row = 4
            col = 0

            btn_h = self.pd_button_height
            btn_y = 0
            self.right_scroll_btn.disable()
            self.left_scroll_btn.disable()

            self.btn_box.visible = False
            for pd in states:
                if btn_h is not None and btn_y is not None and self.y_offset + btn_y + btn_h > self.height:
                    if self._max_button_rows is None:
                        self._max_button_rows = row - 4
                    btn_y = 0
                    row = 4
                    col += 1
                if col in active_cols:
                    pb, btn_h, btn_y = self._make_state_button(pd, row, col)
                    self._state_buttons[pd.tmcc_id] = pb
                else:
                    btn_y += btn_h
                row += 1
            # logic to hide/disable/enable scroll buttons
            if col <= 1:
                self.right_scroll_btn.hide()
                self.left_scroll_btn.hide()
            else:
                if max(active_cols) < col:
                    self.right_scroll_btn.enable()
                if max(active_cols) > 1:
                    self.left_scroll_btn.enable()

            # call post process handler
            self._post_process_state_buttons()
            self.btn_box.visible = True

    def _make_state_button(
        self,
        pd: S | Any,
        row: int,
        col: int,
    ) -> tuple[PushButton | list[Widget], int, int]:
        pb = PushButton(
            self.btn_box,
            text=f"#{pd.tmcc_id} {pd.road_name}",
            grid=[col, row],
            width=int(round(self.width / 2 / (13 * self._scale_by))),
            command=self.switch_state,
            args=[pd],
            padx=0,
            pady=self._button_text_pad_y,
        )
        pb.component_state = pd
        pb.text_size = int(round(15 * self._scale_by))
        pb.bg = self._enabled_bg if self.is_active(pd) else self._disabled_bg
        pb.text_color = self._enabled_text if self.is_active(pd) else self._disabled_text

        # recalculate height
        self.app.update()
        if self.pd_button_width is None:
            self.pd_button_width = pb.tk.winfo_width()
        if self.pd_button_height is None:
            btn_h = self.pd_button_height = pb.tk.winfo_height()
        else:
            btn_h = self.pd_button_height
        btn_y = pb.tk.winfo_y() + btn_h
        return pb, btn_h, btn_y

    def sort_by_number(self) -> None:
        self.by_number.text_bold = True
        self.by_name.text_bold = False

        self.sort_func = lambda x: x.tmcc_id
        states = sorted(self._states.values(), key=self.sort_func)
        self._first_button_col = 0
        self._make_state_buttons(states)

    def sort_by_name(self) -> None:
        self.by_name.text_bold = True
        self.by_number.text_bold = False

        self.sort_func = lambda x: x.road_name.lower()
        states = sorted(self._states.values(), key=self.sort_func)
        self._first_button_col = 0
        self._make_state_buttons(states)

    def scroll_left(self) -> None:
        self._first_button_col -= 1
        states = sorted(self._states.values(), key=self.sort_func)
        self._make_state_buttons(states)

    def scroll_right(self) -> None:
        self._first_button_col += 1
        states = sorted(self._states.values(), key=self.sort_func)
        self._make_state_buttons(states)

    def _post_process_state_buttons(self) -> None:
        pass

    @abstractmethod
    def get_target_states(self) -> list[S]: ...

    @abstractmethod
    def is_active(self, state: S) -> bool: ...

    @abstractmethod
    def switch_state(self, state: S) -> bool: ...


class PowerDistrictsGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        StateBasedGui.__init__(self, "Power Districts", label, width, height, aggrigator, scale_by=scale_by)

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_power_district is True and acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return state.is_aux_on

    def switch_state(self, pd: AccessoryState) -> None:
        with self._cv:
            if pd.is_aux_on:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, pd.tmcc_id).send()


class SwitchesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        StateBasedGui.__init__(self, "Switches", label, width, height, aggrigator, disabled_bg="red", scale_by=scale_by)

    def get_target_states(self) -> list[SwitchState]:
        pds: list[SwitchState] = []
        accs = self._state_store.get_all(CommandScope.SWITCH)
        for acc in accs:
            acc = cast(SwitchState, acc)
            if acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: SwitchState) -> bool:
        return state.is_thru

    def switch_state(self, pd: SwitchState) -> None:
        with self._cv:
            if pd.is_thru:
                CommandReq(TMCC1SwitchCommandEnum.OUT, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1SwitchCommandEnum.THRU, pd.tmcc_id).send()


class RoutesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        StateBasedGui.__init__(self, "Routes", label, width, height, aggrigator, disabled_bg="red", scale_by=scale_by)

    def get_target_states(self) -> list[RouteState]:
        pds: list[RouteState] = []
        accs = self._state_store.get_all(CommandScope.ROUTE)
        for acc in accs:
            acc = cast(RouteState, acc)
            if acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: RouteState) -> bool:
        return state.is_active

    def switch_state(self, pd: RouteState) -> None:
        with self._cv:
            if pd.is_active:
                CommandReq(TMCC1RouteCommandEnum.FIRE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1RouteCommandEnum.FIRE, pd.tmcc_id).send()


class MotorsGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        StateBasedGui.__init__(self, "Motors", label, width, height, aggrigator, scale_by=scale_by)
        self._making_buttons = True

    def _post_process_state_buttons(self) -> None:
        self.app.after(500, self.clear_making_buttons)

    def clear_making_buttons(self) -> None:
        self._making_buttons = False

    @property
    def is_making_buttons(self) -> bool:
        return self._making_buttons

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_amc2 and acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return False

    def switch_state(self, pd: AccessoryState) -> None:
        pass

    # noinspection PyTypeChecker
    def update_button(self, tmcc_id: int) -> None:
        with self._cv:
            pd = self._states[tmcc_id]
            widgets = self._state_buttons[tmcc_id]
            if isinstance(widgets, list):
                for widget in widgets:
                    motor = getattr(widget, "motor", None)
                    lamp = getattr(widget, "lamp", None)
                    if isinstance(widget, PushButton) and motor in {1, 2}:
                        if self.is_motor_active(pd, motor):
                            self._set_button_active(widget)
                        else:
                            self._set_button_inactive(widget)
                    if isinstance(widget, PushButton) and lamp in {1, 2, 3, 4}:
                        if self.is_lamp_active(pd, lamp):
                            self._set_button_active(widget)
                        else:
                            self._set_button_inactive(widget)
                    if isinstance(widget, Slider) and motor in {1, 2}:
                        motor_state = pd.get_motor(motor)
                        if widget.value != motor_state.speed if motor_state else 0:
                            widget.value = motor_state.speed if motor_state else 0.0
                        widget.bg = self._enabled_bg if self.is_motor_active(pd, motor) else "lightgrey"
                    if isinstance(widget, Slider) and lamp in {1, 2, 3, 4}:
                        lamp_state = pd.get_lamp(lamp)
                        if widget.value != lamp_state.level if lamp_state else 0:
                            widget.value = lamp_state.level if lamp_state else 0.0
                        widget.bg = self._enabled_bg if self.is_lamp_active(pd, lamp) else "lightgrey"

    @staticmethod
    def is_motor_active(state: AccessoryState, motor: int) -> bool:
        return state.is_motor_on(state.motor2 if motor == 2 else state.motor1)

    @staticmethod
    def is_lamp_active(state: AccessoryState, lamp: int) -> bool:
        lamp_state = state.get_lamp(lamp)
        return lamp_state and lamp_state.level > 0

    def set_motor_state(self, tmcc_id: int, motor: int, speed: int = None) -> None:
        if self._making_buttons:
            return
        with self._cv:
            pd: AccessoryState = self._states[tmcc_id]
            if speed is None:
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=motor).send()
                if self.is_motor_active(pd, motor):
                    CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, tmcc_id).send()
                else:
                    CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, tmcc_id).send()
            elif speed != (pd.motor1.speed if motor == 1 else pd.motor2.speed):
                Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.MOTOR, motor=motor - 1, speed=speed).send()
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=motor).send()
                if speed:
                    CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, tmcc_id).send()
                else:
                    CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, tmcc_id).send()

    def set_lamp_state(self, tmcc_id: int, lamp: int, level: int = None) -> None:
        if self._making_buttons:
            return
        with self._cv:
            pd: AccessoryState = self._states[tmcc_id]
            lamp_state = pd.get_lamp(lamp)
            if level is None:
                if self.is_lamp_active(pd, lamp):
                    Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.LAMP, lamp=lamp - 1, level=0).send()
                else:
                    Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.LAMP, lamp=lamp - 1, level=100).send()
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=lamp + 2).send()
            elif level != lamp_state.level:
                Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.LAMP, lamp=lamp - 1, level=level).send()
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=lamp + 2).send()

    def _make_state_button(
        self,
        pd: AccessoryState,
        row: int,
        col: int,
    ) -> tuple[list[Widget], int, int]:
        self._making_buttons = True
        ts = int(round(23 * self._scale_by))
        widgets: list[Widget] = []
        # make title label
        title = Text(self.btn_box, text=f"#{pd.tmcc_id} {pd.road_name}", grid=[col, row, 2, 1], size=ts, bold=True)
        widgets.append(title)

        # make motor 1 on/off button
        row += 1
        m1_pwr, btn_h, btn_y = super()._make_state_button(pd, row, col)
        m1_pwr.text = "Motor #1"
        m1_pwr.motor = 1
        if pd.motor1.state:
            self._set_button_active(m1_pwr)
        m1_pwr.update_command(self.set_motor_state, args=[pd.tmcc_id, 1])
        widgets.append(m1_pwr)

        # motor 1 control
        slider_height = int(round(btn_h * 0.9))
        m1_ctl = Slider(
            self.btn_box,
            grid=[col, row + 1],
            height=slider_height,
            width=self.pd_button_width,
            step=5,
        )
        m1_ctl.value = pd.motor1.speed
        m1_ctl.motor = 1
        m1_ctl.bg = self._enabled_bg if pd.motor1.state else "lightgrey"
        m1_db = DebouncedSlider(self, pd.tmcc_id, 1, is_motor=True)
        m1_ctl.update_command(m1_db.on_change)
        widgets.append(m1_ctl)

        # make motor 2 on/off button
        m2_pwr, btn_h, btn_y = super()._make_state_button(pd, row, col + 1)
        m2_pwr.text = "Motor #2"
        m2_pwr.motor = 2
        if self.is_motor_active(pd, 2):
            self._set_button_active(m2_pwr)
        m2_pwr.update_command(self.set_motor_state, args=[pd.tmcc_id, 2])
        widgets.append(m2_pwr)

        # motor 2 control
        m2_ctl = Slider(
            self.btn_box,
            grid=[col + 1, row + 1],
            height=slider_height,
            width=self.pd_button_width,
            step=5,
        )
        m2_ctl.value = pd.motor2.speed
        m2_ctl.motor = 2
        m2_ctl.bg = self._enabled_bg if self.is_motor_active(pd, 2) else "lightgrey"
        m2_db = DebouncedSlider(self, pd.tmcc_id, 2, is_motor=True)
        m2_ctl.update_command(m2_db.on_change)
        widgets.append(m2_ctl)

        # make Lamp controls
        for lamp_no in range(1, 5):
            lamp = pd.get_lamp(lamp_no)
            if lamp_no % 2 == 1:
                row += 2
                lamp_col = col
            else:
                lamp_col = col + 1
            pwr, btn_h, btn_y = super()._make_state_button(pd, row, lamp_col)
            pwr.text = f"Lamp #{lamp_no}"
            pwr.lamp = lamp_no
            if lamp.level:
                self._set_button_active(pwr)
            pwr.update_command(self.set_lamp_state, args=[pd.tmcc_id, lamp_no])
            widgets.append(pwr)

            slider_height = int(round(btn_h * 0.9))
            ctl = Slider(
                self.btn_box,
                grid=[lamp_col, row + 1],
                height=slider_height,
                width=self.pd_button_width,
                step=5,
            )
            ctl.value = lamp.level
            ctl.lamp = lamp_no
            ctl.bg = self._enabled_bg if lamp.level else "lightgrey"
            dbs = DebouncedSlider(self, pd.tmcc_id, lamp_no, is_lamp=True)
            ctl.update_command(dbs.on_change)
            widgets.append(ctl)

        # noinspection PyTypeChecker
        self._state_buttons[pd.tmcc_id] = widgets
        return widgets, btn_h, btn_y


class AccessoriesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        self._is_momentary = set()
        self._released_events = dict[int, Event]()
        StateBasedGui.__init__(self, "Accessories", label, width, height, aggrigator, scale_by=scale_by)

    def _post_process_state_buttons(self) -> None:
        for tmcc_id in self._is_momentary:
            if tmcc_id in self._state_buttons:
                pb = self._state_buttons[tmcc_id]
                pb.when_left_button_pressed = self.when_pressed
                pb.when_left_button_released = self.when_released

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_power_district or acc.is_sensor_track or acc.is_amc2:
                continue
            if acc.road_name and acc.road_name.lower().strip() != "unused":
                pds.append(acc)
                name_lc = acc.road_name.lower()
                if "aux1" in name_lc or "ax1" in name_lc or "(a1)" in name_lc or "(m)" in name_lc:
                    self._is_momentary.add(acc.address)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return state.is_aux_on

    def switch_state(self, pd: AccessoryState) -> None:
        with self._cv:
            if pd.tmcc_id in self._is_momentary:
                pass
            elif pd.is_aux_on:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()

    def when_pressed(self, event: EventData) -> None:
        pb = event.widget
        state = pb.component_state
        if state.is_asc2:
            Asc2Req(state.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1).send()
        else:
            if state.tmcc_id in self._released_events:
                event = self._released_events[state.tmcc_id]
                self._released_events[state.tmcc_id].clear()
            else:
                self._released_events[state.tmcc_id] = event = Event()
            _ = MomentaryActionHandler(pb, event, state, 0.2)
        # self.app.after(10, self._set_button_active, [event.widget])

    def when_released(self, event: EventData) -> None:
        state = event.widget.component_state
        if state.is_asc2:
            Asc2Req(state.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=0).send()
        else:
            state = event.widget.component_state
            self._released_events[state.tmcc_id].set()
        # self.app.after(10, self._set_button_inactive, [event.widget])


class ComponentStateGui(Thread):
    def __init__(
        self,
        label: str = None,
        initial: str = "Power Districts",
        width: int = None,
        height: int = None,
        scale_by: float = 1.0,
    ) -> None:
        super().__init__(daemon=True)
        self._ev = Event()
        self._guis = {
            "Accessories": AccessoriesGui,
            "Motors": MotorsGui,
            "Power Districts": PowerDistrictsGui,
            "Routes": RoutesGui,
            "Switches": SwitchesGui,
        }
        # verify requested GUI exists:
        if initial not in self._guis:
            raise ValueError(f"Invalid initial GUI: {initial}")

        self.label = label
        if width is None or height is None:
            try:
                from tkinter import Tk

                root = Tk()
                self.width = root.winfo_screenwidth()
                self.height = root.winfo_screenheight()
                root.destroy()
            except Exception as e:
                log.exception("Error determining window size", exc_info=e)
        else:
            self.width = width
            self.height = height
        self._scale_by = scale_by
        self._gui = None
        self.requested_gui = initial

        self.start()

    def run(self) -> None:
        # create the initially requested gui
        self._gui = self._guis[self.requested_gui](
            self.label, self.width, self.height, aggrigator=self, scale_by=self._scale_by
        )

        # wait for user to request a different GUI
        while True:
            # Wait for request to change GUI
            self._ev.wait()
            self._ev.clear()

            # Close/destroy previous GUI
            GpioHandler.release_handler(self._gui)

            # wait for Gui to be destroyed
            self._gui.destroy_complete.wait(10)
            self._gui.join()
            # clean up state
            self._gui = None

            # create and display new gui
            self._gui = self._guis[self.requested_gui](
                self.label, self.width, self.height, aggrigator=self, scale_by=self._scale_by
            )

    def cycle_gui(self, gui: str):
        if gui in self._guis:
            self.requested_gui = gui
            self._ev.set()

    @property
    def guis(self) -> list[str]:
        return list(self._guis.keys())


class DebouncedSlider:
    def __init__(
        self,
        gui: MotorsGui,
        tmcc_id: int,
        device: int,
        delay_ms=500,
        is_motor: bool = False,
        is_lamp: bool = False,
    ) -> None:
        self._is_lamp = is_lamp
        self._is_motor = is_motor
        self._gui = gui
        self._tmcc_id = tmcc_id
        self._device = device
        self._app = gui.app
        self._tk = gui.app.tk
        self._after_id = None
        self._delay = delay_ms
        self._last_value = None

    def on_change(self, value):
        if self._gui.is_making_buttons:
            return
        self._last_value = int(value)
        # Cancel previously scheduled call if any
        if self._after_id is not None:
            self._tk.after_cancel(self._after_id)
            self._after_id = None
        # Schedule new call after user stops moving for delay_ms
        self._after_id = self._tk.after(self._delay, self._fire)

    def _fire(self):
        self._after_id = None
        self.commit(self._last_value)

    def commit(self, value):
        # Do the real work once sliding has paused
        if self._is_motor:
            self._gui.set_motor_state(self._tmcc_id, self._device, value)
        elif self._is_lamp:
            self._gui.set_lamp_state(self._tmcc_id, self._device, value)


class MomentaryActionHandler(Thread, Generic[S]):
    def __init__(self, widget: PushButton, event: Event, state: S, timeout: float) -> None:
        super().__init__(daemon=True)
        self._widget = widget
        self._ev = event
        self._state = state
        self._timeout = timeout
        self.start()

    def run(self) -> None:
        while not self._ev.wait(self._timeout):
            if not self._ev.is_set():
                print("still pressed")
            else:
                break



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
import pytest

from src.pytrain import TMCC2EffectsControl
from src.pytrain.db.comp_data import LEGACY_CONTROL_TYPE, CompData
from src.pytrain.db.engine_state import EngineState, TrainState
from src.pytrain.pdi.base3_component import ConsistComponent
from src.pytrain.protocol.constants import CommandScope
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum as TMCC1
from src.pytrain.protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum as TMCC2


class TestEngineStateBehavior:
    @staticmethod
    def _new_engine(addr: int = 7) -> EngineState:
        st = EngineState(CommandScope.ENGINE)
        # Simulate that component data is known/allocated
        st.initialize(CommandScope.ENGINE, addr)
        # Set address so properties relying on it work predictably
        st._address = addr  # type: ignore[attr-defined]
        return st

    def test_is_legacy_deduction_by_address_and_control_type(self):
        e1 = self._new_engine(addr=7)
        # No control type set, address <= 99 defaults to TMCC unless control_type marks Legacy
        assert e1.is_legacy is False

        # Force Legacy via control_type on base data
        e1.comp_data._control_type = LEGACY_CONTROL_TYPE  # type: ignore[attr-defined]
        assert e1.is_legacy is True

        # 4-digit address implies Legacy if _is_legacy not explicitly set
        e2 = self._new_engine(addr=2345)
        # Clear any control_type so address rule applies
        e2.comp_data._control_type = None  # type: ignore[attr-defined]
        assert e2.is_legacy is True

    def test_decode_speed_info_255_conversion(self):
        e = self._new_engine()
        # Explicitly force protocol flavor
        assert not e.is_legacy
        assert e.decode_speed_info(255) == 31
        e._is_legacy = True  # type: ignore[attr-defined]
        assert e.is_legacy
        assert e.decode_speed_info(255) == 195
        # Pass-through non-255
        assert e.decode_speed_info(20) == 20

    def test_speed_max_logic_limits(self):
        e = self._new_engine()
        e._is_legacy = True  # type: ignore[attr-defined]
        e.comp_data._speed_limit = 255  # type: ignore[attr-defined]
        assert e.speed_max == 199  # legacy default cap when limits unset (255)

        e.comp_data._max_speed = 200  # type: ignore[attr-defined]
        e.comp_data._speed_limit = 255  # type: ignore[attr-defined]
        assert e.speed_max == 200  # legacy default cap when limits unset (255)

        # Non-legacy should never exceed 31
        e._is_legacy = False  # type: ignore[attr-defined]
        e.comp_data._max_speed = 100  # type: ignore[attr-defined]
        e.comp_data._speed_limit = 100  # type: ignore[attr-defined]
        assert e.speed_max == 31

        # Both limits present -> pick min; still capped for TMCC
        e.comp_data._max_speed = 20  # type: ignore[attr-defined]
        e.comp_data._speed_limit = 25  # type: ignore[attr-defined]
        assert e.speed_max == 20

    def test_change_direction_toggle_tmcc1_and_tmcc2(self):
        e = self._new_engine()
        # Start in TMCC1; set a TMCC1 direction
        e._is_legacy = False  # type: ignore[attr-defined]
        e._direction = TMCC1.FORWARD_DIRECTION  # type: ignore[attr-defined]
        nd = e._change_direction(TMCC1.TOGGLE_DIRECTION)
        assert nd == TMCC1.REVERSE_DIRECTION

        nd2 = e._change_direction(TMCC1.TOGGLE_DIRECTION)
        # _change_direction doesn't mutate state; simulate applying result
        e._direction = nd2  # type: ignore[attr-defined]
        nd3 = e._change_direction(TMCC1.TOGGLE_DIRECTION)
        assert nd3 == TMCC1.FORWARD_DIRECTION

        # Now simulate Legacy/TMCC2 flavor and TMCC2 direction
        e._is_legacy = True  # type: ignore[attr-defined]
        e._direction = TMCC2.FORWARD_DIRECTION  # type: ignore[attr-defined]
        nd4 = e._change_direction(TMCC2.TOGGLE_DIRECTION)
        assert nd4 == TMCC2.REVERSE_DIRECTION

    def test_as_bytes_yields_packets_list(self):
        e = self._new_engine(addr=12)  # <=99 so BaseReq BASE_MEMORY should be first
        # Provide some extra state that can be serialized if present
        e._start_stop = TMCC1.START_UP_IMMEDIATE  # type: ignore[attr-defined]
        # Use a smoke value the mapping will understand
        e.comp_data.smoke_tmcc = TMCC2EffectsControl.SMOKE_LOW

        packets = e.as_bytes()
        # EngineState.as_bytes returns list[bytes]
        assert isinstance(packets, list)
        assert len(packets) >= 1
        assert all(isinstance(p, (bytes, bytearray)) for p in packets)

    def test_smoke_label_mapping(self):
        e = self._new_engine()
        assert e.is_tmcc

        # force legacy mode
        e._comp_data._control_type = LEGACY_CONTROL_TYPE
        assert e.is_legacy
        e.comp_data.smoke_tmcc = TMCC2EffectsControl.SMOKE_OFF
        assert e.smoke_label == "-"
        e.comp_data.smoke_tmcc = TMCC2EffectsControl.SMOKE_LOW
        assert e.smoke_label == "L"
        e.comp_data.smoke_tmcc = TMCC2EffectsControl.SMOKE_MEDIUM
        assert e.smoke_label == "M"
        e.comp_data.smoke_tmcc = TMCC2EffectsControl.SMOKE_HIGH
        assert e.smoke_label == "H"

    def test_as_dict_includes_core_fields_engine(self):
        e = self._new_engine()
        # Populate some comp data fields
        e.comp_data._speed = 10  # type: ignore[attr-defined]
        e.comp_data._momentum = 73  # type: ignore[attr-defined]
        e.comp_data._rpm_labor = CompData.encode_rpm_labor(rpm=5, labor=14)  # type: ignore[attr-defined]
        e.comp_data._engine_type = 0  # type: ignore[attr-defined]
        e.comp_data._sound_type = 0  # type: ignore[attr-defined]
        e.comp_data._engine_class = 0  # type: ignore[attr-defined]
        e.comp_data._control_type = LEGACY_CONTROL_TYPE  # type: ignore[attr-defined]
        e._direction = TMCC2.FORWARD_DIRECTION  # type: ignore[attr-defined]
        e.comp_data.smoke_tmcc = TMCC2EffectsControl.SMOKE_LOW

        d = e.as_dict()
        # spot-check expected keys
        assert d["scope"] == "engine"
        assert d["tmcc_id"] == e.tmcc_id
        assert d["speed"] == 10
        assert d["direction"] == "forward_direction"
        assert d["smoke"] == "smoke_low"
        assert "engine_type" in d and "sound_type" in d and "engine_class" in d

    def test_can_not_define_out_of_scope_field(self):
        e = self._new_engine()
        with pytest.raises(AttributeError):
            _ = e.this_attribute_does_not_exist

    def test_train_state_as_dict_includes_consist(self):
        t = TrainState(CommandScope.TRAIN)
        t.initialize(CommandScope.TRAIN, tmcc_id=101)
        t._address = 101  # type: ignore[attr-defined]
        # Add some consist components
        t.comp_data._consist_flags = 0b10101010  # type: ignore[attr-defined]
        t.comp_data._consist_comps = [  # type: ignore[attr-defined]
            ConsistComponent(tmcc_id=11, flags=0b00000001),
            ConsistComponent(tmcc_id=22, flags=0b00000100),
        ]
        d = t.as_dict()
        assert d["scope"] == "train"
        assert isinstance(d["flags"], int)
        assert isinstance(d["components"], dict)
        # Keys are component tmcc_ids; values are info strings
        assert 11 in d["components"] and 22 in d["components"]



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
from typing import cast

# Python
import pytest

from src.pytrain.pdi.constants import (
    ALL_FIRMWARE,
    ALL_IDENTIFY,
    ALL_INFO,
    ALL_STATUS,
    PDI_EOP,
    PDI_SOP,
    ALL_SETs,
    PdiAction,
    PdiCommand,
    Ser2Action,
)
from src.pytrain.pdi.lcs_req import ERROR_CODE_MAP, LcsReq, Ser2Req
from src.pytrain.pdi.pdi_req import PdiReq


# noinspection PyProtectedMember
def build_lcs_packet(
    pdi_cmd: PdiCommand, action: PdiAction, tmcc_id: int, payload: bytes = b"", is_error=False
) -> bytes:
    # [SOP][PDI command(2)][tmcc id(1)][action(1)][payload...][EOP]
    # action with MSB set indicates error
    if is_error:
        act = action.value.bits | 0x80
    else:
        act = action.value.bits
    inner = b"".join(
        [
            pdi_cmd.as_bytes,
            tmcc_id.to_bytes(1, "big"),
            act.to_bytes(1, "big"),
            payload,
        ]
    )
    stuffed, checksum = PdiReq._calculate_checksum(inner, add_stf=True)  # noqa: SLF001
    return b"".join([PDI_SOP.to_bytes(1, "big"), stuffed, checksum, PDI_EOP.to_bytes(1, "big")])


def test_is_lcs_true_and_repr_safe_without_payload():
    req = Ser2Req(7, PdiCommand.SER2_GET, Ser2Action.CONFIG)
    assert req.is_lcs is True
    s = repr(req)
    assert "PDI" in s and "ID: 7" in s


def test_parse_status_from_bytes_populates_fields_and_payload_text():
    # payload fields per LcsReq for ALL_STATUS
    # [board_id, num_ids, model, uart0, uart1, base_type, dc_volts*10, ec0..ec15]
    board_id = 2
    num_ids = 5
    model = 9
    uart0 = 2
    uart1 = 3
    base_type = 1  # "Legacy"
    dc_volts_tenths = 178  # -> 17.8V
    errs = bytes(range(0x10, 0x10 + 16))  # ec0..ec15 diverse data
    payload = bytes([board_id, num_ids, model, uart0, uart1, base_type, dc_volts_tenths]) + errs

    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.STATUS, 25, payload)
    parsed = PdiReq.from_bytes(pkt)
    assert isinstance(parsed, Ser2Req)

    # action mapping
    assert parsed.action == Ser2Action.STATUS

    # fields
    assert parsed.board_id == board_id
    assert parsed.num_ids == num_ids
    assert parsed.model == model
    assert parsed.uart0 == uart0
    assert parsed.uart1 == uart1
    assert parsed.base_type == "Legacy"
    assert pytest.approx(parsed.dc_volts, rel=0, abs=1e-6) == dc_volts_tenths / 10.0

    # payload (human text) should include key tokens and packet echo
    s = parsed.payload
    assert "Board ID:" in s and "Num IDs:" in s and "Model:" in s and "DC Volts:" in s
    assert "UART0:" in s and "UART1:" in s and "Base:" in s
    # ensure UART mode mapping text appears
    assert ">Base<" in s or "->Base" in s or "Base->" in s


def test_parse_firmware_payload_text():
    version, revision, sub_rev = 1, 2, 3
    payload = bytes([version, revision, sub_rev])
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.FIRMWARE, 3, payload)
    parsed = PdiReq.from_bytes(pkt)

    assert isinstance(parsed, Ser2Req)
    assert parsed.action == Ser2Action.FIRMWARE
    assert parsed.payload == f"Firmware {version}.{revision}.{sub_rev}"


def test_parse_info_payload_text():
    board_id = 7
    num_ids = 42
    model = 11
    dc_volts_tenths = 129
    payload = bytes([board_id, num_ids, model, dc_volts_tenths])
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.INFO, 77, payload)
    parsed = PdiReq.from_bytes(pkt)

    assert isinstance(parsed, Ser2Req)
    assert parsed.action == Ser2Action.INFO
    s = parsed.payload
    assert "Board ID:" in s and "Num IDs:" in s and "Model:" in s
    assert "DC Volts:" in s
    assert str(dc_volts_tenths / 10.0) in s


def test_error_flag_sets_is_error_and_maps_error_code_text():
    # Action with MSB set indicates error; error code is first payload byte
    error_code = 2  # "Action not supported"
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.CONFIG, 9, bytes([error_code]), is_error=True)
    parsed = PdiReq.from_bytes(pkt)

    assert isinstance(parsed, Ser2Req)
    assert parsed.is_error is True
    assert "Action not supported" in parsed.error
    # payload falls back to base PdiReq payload when error
    assert isinstance(parsed.payload, str)


def test_error_code_unknown_maps_to_none_text():
    error_code = 0x00  # supported; use an unmapped like 0 to force "None"
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.CONFIG, 9, bytes([error_code]), is_error=True)
    parsed = cast(LcsReq, PdiReq.from_bytes(pkt))
    assert parsed.error == "None"


def test_error_code_error_texts():
    for error_code, error_text in ERROR_CODE_MAP.items():
        pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.CONFIG, 9, bytes([error_code]), is_error=True)
        parsed = cast(LcsReq, PdiReq.from_bytes(pkt))
        assert parsed.error == error_text


def test_uart_mode_mapping_and_base_type_defaults_to_na():
    # Use a non-mapped base_type and uart values
    payload = bytes([0, 0, 0, 99, 100, 99, 0])
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.STATUS, 1, payload)
    parsed = cast(LcsReq, PdiReq.from_bytes(pkt))
    assert parsed.base_type == "NA"
    assert LcsReq.uart_mode(99) == "NA"
    assert LcsReq.uart_mode(1) == ">Base<"


def test_identify_set_payload_includes_ident_when_command_is_set():
    # Only ALL_IDENTIFY with a command in ALL_SETs returns "Ident: {ident} ..."
    # Build a request object (not from bytes) so action gate evaluates AND command name endswith _SET
    ident = 0x5A
    req = Ser2Req(12, PdiCommand.SER2_SET, Ser2Action.IDENTIFY, ident=ident)
    assert Ser2Action.IDENTIFY in ALL_IDENTIFY
    assert req.pdi_command in ALL_SETs
    assert "Ident: 90" in req.payload  # 0x5A == 90


def test_payload_falls_back_when_not_matching_any_clause():
    # Build a GET CONFIG where payload should fall back unless special formatting defined
    req = Ser2Req(2, PdiCommand.SER2_GET, Ser2Action.CONFIG)
    s = req.payload
    # base PdiReq payload is a string or None; ensure not crashing and is str
    assert s is None or isinstance(s, str)


def test_ser2req_from_bytes_sets_action_enum():
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.CONFIG, 5)
    parsed = PdiReq.from_bytes(pkt)
    assert isinstance(parsed, Ser2Req)
    assert parsed.action == Ser2Action.CONFIG


def test_is_action_helper_against_lists():
    # Ensure _is_action recognizes membership for different lists
    # Build minimal object via bytes to exercise internal path
    pkt = build_lcs_packet(PdiCommand.SER2_RX, Ser2Action.STATUS, 9, b"\x00\x00\x00\x00\x00\x00\x00")
    parsed = PdiReq.from_bytes(pkt)
    assert isinstance(parsed, Ser2Req)
    assert parsed.action in ALL_STATUS
    assert parsed.action not in ALL_FIRMWARE
    assert parsed.action not in ALL_INFO
    assert parsed.action not in ALL_IDENTIFY



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import types
from typing import Any, Optional

import pytest

from src.pytrain.protocol.constants import CommandScope


# noinspection PyUnusedLocal
class DummyOled:
    def __init__(self, address: Optional[int], device: Any, auto_update: bool = False, scroll_rate: float = 0.03):
        self._rows = ["", "", "", ""]
        self._cols = 21  # pick width > 20 to exercise a wider formatting branch
        self._updates = 0
        self._clears = 0
        self._writes = []  # records (text, row, kwargs)
        self._cursor_pos = None

    def __setitem__(self, index: int, value: str) -> None:
        self._rows[index] = value

    def __getitem__(self, index: int) -> str:
        return self._rows[index]

    @property
    def rows(self) -> int:
        return len(self._rows)

    @property
    def cols(self) -> int:
        return self._cols

    @cols.setter
    def cols(self, v: int) -> None:
        self._cols = v

    def write(self, text: str, row: int, **kwargs):
        # record writes and place centered text in buffer if requested
        self._writes.append((text, row, kwargs))
        center = kwargs.get("center", False)
        blink = kwargs.get("blink", False)
        if center:
            padding = max(0, (self._cols - len(text)) // 2)
            s = " " * padding + text
        else:
            s = text
        if blink:
            # just annotate blink in buffer minimally for assertion
            s = s  # keep same; we rely on _writes to assert blinking prompt
        self._rows[row] = s[: self._cols]

    def clear(self):
        self._clears += 1
        self._rows = ["", "", "", ""]

    def update_display(self):
        self._updates += 1

    def reset(self):
        # emulate turning off/cleanup
        pass

    def show(self):
        pass

    def hide(self):
        pass

    @property
    def cursor_pos(self):
        return self._cursor_pos

    @cursor_pos.setter
    def cursor_pos(self, value):
        self._cursor_pos = value


class DummyWatcher:
    def __init__(self, state, action):
        self.state = state
        self.action = action
        self._shutdown = False

    def shutdown(self):
        self._shutdown = True


class DummyState:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return f"DummyState({self.__dict__})"


# noinspection PyUnusedLocal
class DummyStateStore:
    def __init__(self):
        self.states = {}

    def set_state(self, scope: CommandScope, address: int, state: Any):
        self.states[(scope, address)] = state

    def get_state(self, scope: CommandScope, address: int, create: bool = True):
        return self.states.get((scope, address))


@pytest.fixture()
def patched_engine_status(monkeypatch):
    # Import target after patching module-level dependencies
    import src.pytrain.gpio.engine_status as es_mod

    # Patch atexit to avoid global registrations
    monkeypatch.setattr(es_mod.atexit, "register", lambda *_args, **_kw: None, raising=True)

    # Patch Oled and StateWatcher in the module under test
    monkeypatch.setattr(es_mod, "Oled", DummyOled, raising=True)
    monkeypatch.setattr(es_mod, "StateWatcher", DummyWatcher, raising=True)

    # Prepare and patch a stub state store
    store = DummyStateStore()
    # SYNC state: address 99
    sync_state = DummyState(is_synchronized=False)
    store.set_state(CommandScope.SYNC, 99, sync_state)
    # BASE state: used for railroad name
    base_state = DummyState(base_name="my railroad")
    store.set_state(CommandScope.BASE, 0, base_state)

    # Patch ComponentStateStore.get() to return our dummy store
    class CSStoreProxy:
        @staticmethod
        def get():
            return store

    monkeypatch.setattr(es_mod, "ComponentStateStore", CSStoreProxy, raising=True)

    # Provide convenience access in the fixture
    return types.SimpleNamespace(module=es_mod, store=store, sync_state=sync_state)


def make_engine_state(
    address: int = 1234,
    scope: CommandScope = CommandScope.ENGINE,
    overrides: Optional[dict] = None,
):
    defaults = dict(
        address=address,
        scope=scope,
        is_started=True,
        is_shutdown=False,
        road_name="My Engine",
        road_number=str(address),
        control_type_label="Legacy",
        engine_type_label="Steam",
        speed=12,
        speed_max=200,
        direction_label="FW",
        rpm=1,
        labor=7,
        train_brake=0,
        momentum=3,
        smoke_label="Med",
    )
    if overrides:
        defaults.update(overrides)
    return DummyState(**defaults)


def test_initial_sync_incomplete_shows_synchronizing_and_sets_watcher(patched_engine_status):
    es_mod = patched_engine_status.module
    # Initial sync state is false per fixture
    eng_status = es_mod.EngineStatus(tmcc_id=1234, scope=CommandScope.ENGINE)

    # Should have written "Synchronizing..." with blink=True on row 0 via DummyOled.write
    writes = eng_status.display._writes
    assert any(w[0] == "Synchronizing..." and w[1] == 0 and w[2].get("blink") is True for w in writes), (
        "Expected synchronizing prompt with blink"
    )

    # A sync watcher should have been created
    assert isinstance(eng_status._sync_watcher, DummyWatcher)
    # No monitored engine state yet since store has none
    assert eng_status.state is None

    # cleanup
    eng_status.reset()


def test_on_sync_sets_synchronized_and_updates_display_not_found(patched_engine_status):
    es_mod = patched_engine_status.module
    eng_status = es_mod.EngineStatus(tmcc_id=2222, scope=CommandScope.ENGINE)

    # Flip sync state to True and trigger on_sync
    patched_engine_status.sync_state.is_synchronized = True
    eng_status.on_sync()

    # Sync watcher is shut down and cleared
    assert eng_status._sync_watcher is None
    assert eng_status.is_synchronized is True

    # No engine state present in store for tmcc_id=2222 => should show Not Found
    writes = eng_status.display._writes
    assert any("Not Found" in w[0] and w[1] == 2 for w in writes), "Expected 'Not Found' on row 2"

    # Should have cleared display once during update_display(clear=True)
    assert eng_status.display._clears >= 1

    eng_status.reset()


def test_update_engine_sets_state_and_clears_display_and_starts_watcher(patched_engine_status):
    es_mod = patched_engine_status.module
    # Start synchronized so EngineStatus won't create sync watcher that interferes
    patched_engine_status.sync_state.is_synchronized = True

    # Preload an engine state in store
    state = make_engine_state(address=3333)
    patched_engine_status.store.set_state(CommandScope.ENGINE, 3333, state)

    eng_status = es_mod.EngineStatus(tmcc_id=3333, scope=CommandScope.ENGINE)

    # Engine state watcher created
    assert isinstance(eng_status._state_watcher, DummyWatcher)
    assert eng_status.state is state

    # Trigger update_engine to another ID with no state to ensure clear is called
    eng_status.update_engine(4444, CommandScope.ENGINE)
    assert eng_status.tmcc_id == 4444
    assert eng_status.scope == CommandScope.ENGINE
    assert eng_status.display._clears >= 1
    # Since 4444 not in store, monitored state should be None
    assert eng_status.state is None

    eng_status.reset()


def test_update_display_with_state_populates_rows_and_cursor(patched_engine_status):
    es_mod = patched_engine_status.module
    patched_engine_status.sync_state.is_synchronized = True

    # Provide monitored state directly via constructor by passing state object
    eng_state = make_engine_state(address=5555)
    patched_engine_status.store.set_state(CommandScope.ENGINE, 5555, eng_state)
    eng_status = es_mod.EngineStatus(tmcc_id=getattr(eng_state, "address"), scope=CommandScope.ENGINE)

    # After init, update_display is invoked via on_sync path or direct; ensure rows are set
    # Row 0 contains road info and labels
    assert "My Engine" in eng_status.display[0]
    assert "Legacy" in eng_status.display[0] or "Steam" in eng_status.display[0]

    # Row 1 contains scope label and tmcc id
    assert f"{eng_status.scope.label}:" in eng_status.display[1]
    assert f"{eng_status.tmcc_id:04}" in eng_status.display[1]

    # Row 2 contains Speed and direction text adapted for cols > 20
    assert "Speed:" in eng_status.display[2]
    assert "Fwd" in eng_status.display[2] or "Rev" in eng_status.display[2] or "---" in eng_status.display[2]

    # Row 3 contains TB, Mo, Sm, and RPM when cols > 20
    assert "TB:" in eng_status.display[3]
    assert "Mo:" in eng_status.display[3]
    assert "Sm:" in eng_status.display[3]
    assert "RPM:" in eng_status.display[3]

    # Cursor expected at (1, 8) when a monitored state is present
    assert eng_status.display.cursor_pos == (1, 8)

    eng_status.reset()



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import re
import types

import pytest

from src.pytrain.db.comp_data import (
    AccessoryData,
    CompData,
    CompDataHandler,
    CompDataMixin,
    EngineData,
    RouteData,
    SwitchData,
    TrainData,
    UpdatePkg,
)
from src.pytrain.pdi.pdi_req import PdiReq
from src.pytrain.protocol.constants import LEGACY_CONTROL_TYPE, CommandScope
from src.pytrain.protocol.multibyte.multibyte_constants import TMCC2EffectsControl
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum


class TestCompData:
    def test_comp_data_handler(self):
        # default handler
        h = CompDataHandler("_speed")
        assert h.field == "_speed"
        assert h.length == 1
        assert callable(h.from_bytes)
        assert callable(h.to_bytes)
        assert h.is_d4_only is False

        # custom handler (multibyte, d4_only)
        to_bytes_called = []

        def _to_bytes(x: int) -> bytes:
            to_bytes_called.append(True)
            return x.to_bytes(2, "little")

        h2 = CompDataHandler("_bt_id", length=2, to_bytes=_to_bytes, d4_only=True)
        assert h2.field == "_bt_id"
        assert h2.length == 2
        assert h2.is_d4_only is True
        # smoke test conversion callback
        bts = h2.to_bytes(513)  # 0x0201
        assert bts == b"\x01\x02"
        assert to_bytes_called, "custom to_bytes must be invoked"

    def test_update_pkg(self):
        pkg = UpdatePkg("speed", 0x07, 1, b"\x2a")
        s = repr(pkg)
        # format: "<field>: Address: <hex> Length: <len> data: <hex data>"
        assert "speed" in s
        assert "Address: 0x7" in s
        assert "Length: 1" in s
        assert re.search(r"data:\s*2a", s) is not None

    def test_comp_data_from_bytes_factory(self):
        # Build a buffer of correct length for each scope and ensure subclass is returned
        for scope, expected_cls in [
            (CommandScope.ENGINE, EngineData),
            (CommandScope.TRAIN, TrainData),
            (CommandScope.ACC, AccessoryData),
            (CommandScope.SWITCH, SwitchData),
            (CommandScope.ROUTE, RouteData),
        ]:
            buf = b"\xff" * PdiReq.scope_record_length(scope)
            tmcc_id = 1234 if scope in {CommandScope.ENGINE, CommandScope.TRAIN} else 12
            obj = CompData.from_bytes(buf, scope, tmcc_id=tmcc_id)
            assert isinstance(obj, expected_cls)
            assert obj._scope == scope  # internal scope set by base class

        with pytest.raises(ValueError):
            # Invalid scope must raise
            CompData.from_bytes(b"", types.SimpleNamespace(name="BOGUS"), tmcc_id=1)  # type: ignore[arg-type]

    def test_engine_data_is_legacy_flag_and_smoke_mapping(self):
        # Initialize with padded bytes so all fields get some numeric defaults
        buf = b"\xff" * PdiReq.scope_record_length(CommandScope.ENGINE)
        eng = EngineData(buf, tmcc_id=1000)

        # Default: not legacy unless control_type matches constant
        assert eng.is_legacy is False
        # Force legacy and test TMCC2 smoke mapping via _tmcc attribute access
        eng._control_type = LEGACY_CONTROL_TYPE

        # Set raw base smoke value (0..3) and read TMCC2 enum via property mapping
        eng.smoke = 3  # sets _smoke directly via overridden __setattr__
        assert eng.smoke_tmcc == TMCC2EffectsControl.SMOKE_HIGH

        # Now flip to non-legacy mapping (TMCC1). Reading should map to TMCC1 enums.
        eng._control_type = 0x00  # anything different
        eng.smoke = 0
        assert eng.smoke_tmcc == TMCC1EngineCommandEnum.SMOKE_OFF

        # Setting smoke via TMCC mapping should store base value correctly
        eng._control_type = LEGACY_CONTROL_TYPE  # legacy again => use TMCC2 -> base map
        eng.smoke_tmcc = TMCC2EffectsControl.SMOKE_MEDIUM
        assert eng.smoke == 2

        eng._control_type = 0x00  # non-legacy => TMCC1 -> base map
        eng.smoke_tmcc = TMCC1EngineCommandEnum.SMOKE_ON
        assert eng.smoke == 1

    def test_engine_data_conversions_basic(self):
        buf = b"\xff" * PdiReq.scope_record_length(CommandScope.ENGINE)
        eng = EngineData(buf, tmcc_id=1000)

        # momentum conversion: external "tmcc" value <-> base value
        eng.momentum_tmcc = 6
        # base value should be roughly round(6 * 18.14) = 3 (capped 0..127)
        assert eng.momentum == 109
        # reading back should invert: round(109 * 0.05512) -> ~54 (capped 0..7)
        assert eng.momentum_tmcc == 6

        # train_brake conversion
        eng.train_brake_tmcc = 5
        # base ~ round(5 * 2.143)=11
        assert eng.train_brake == 11
        # invert
        assert eng.train_brake_tmcc == 5

        # rpm/labor combined conversion
        eng.rpm_tmcc = 5
        eng.labor_tmcc = 14
        # verify combined field was encoded
        assert isinstance(eng._rpm_labor, int)
        # accessing rpm/labor again returns masked/decoded
        assert eng.rpm_tmcc == 5
        assert eng.labor_tmcc == 14

    def test_repr_formats_by_scope(self):
        # Engine/Train use 4 digits, others 2, include payload, name, and number
        ebuf = b"\xff" * PdiReq.scope_record_length(CommandScope.ENGINE)
        eng = EngineData(ebuf, tmcc_id=7)
        eng.road_name = "abc railroad"
        eng.road_number = "123"
        s_eng = repr(eng)
        assert "Engine 0007:" in s_eng
        assert "#123" in s_eng
        # Title-cased road name
        assert "ABC Railroad" in s_eng

        sbuf = b"\xff" * PdiReq.scope_record_length(CommandScope.SWITCH)
        sw = SwitchData(sbuf, tmcc_id=7)
        sw.road_name = "yard"
        sw.road_number = "5"
        s_sw = repr(sw)
        assert "Switch  7:" in s_sw
        assert "Yard" in s_sw
        assert "#5" in s_sw

    def test_as_bytes_round_length_and_padding(self):
        # For each scope, ensure as_bytes returns a buffer sized to scope_record_length
        for scope, cls in [
            (CommandScope.ENGINE, EngineData),
            (CommandScope.TRAIN, TrainData),
            (CommandScope.ACC, AccessoryData),
            (CommandScope.SWITCH, SwitchData),
            (CommandScope.ROUTE, RouteData),
        ]:
            data_len = PdiReq.scope_record_length(scope)
            b = b"\xff" * data_len
            # tmcc_id: >99 for engine/train to keep 4-digit fields present, <=99 for others
            tmcc_id = 1234 if scope in {CommandScope.ENGINE, CommandScope.TRAIN} else 12
            obj = cls(b, tmcc_id=tmcc_id)  # type: ignore[call-arg]
            out = obj.as_bytes()
            assert isinstance(out, bytes)
            assert len(out) == data_len

    def test_route_payload_renders_components(self):
        # Build a RouteData and inject simple components with the required attributes
        buf = b"\xff" * PdiReq.scope_record_length(CommandScope.ROUTE)
        route = RouteData(buf, tmcc_id=42)

        class _FakeSwitch:
            is_route = False
            is_switch = True

            def __init__(self, tid: int, thru: bool):
                self.tmcc_id = tid
                self.is_thru = thru

        class _FakeRoute:
            is_route = True
            is_switch = False

            def __init__(self, tid: int):
                self.tmcc_id = tid

        # Set via CompData __setattr__ (non-underscore attribute supported)
        route.components = [
            _FakeRoute(3),
            _FakeSwitch(7, True),
            _FakeSwitch(9, False),
            _FakeRoute(11),
        ]
        payload = route.payload()
        # Expected fragments appear
        assert "Routes:" in payload and " 3" in payload and "11" in payload
        assert "Switches:" in payload and " 7 [thru]" in payload and " 9 [out]" in payload

    def test_comp_data_mixin_initialize_sets_types_and_flag(self):
        class Holder(CompDataMixin):
            def __init__(self):
                super().__init__()

        for scope, expected_cls in [
            (CommandScope.ENGINE, EngineData),
            (CommandScope.TRAIN, TrainData),
            (CommandScope.ACC, AccessoryData),
            (CommandScope.SWITCH, SwitchData),
            (CommandScope.ROUTE, RouteData),
        ]:
            h = Holder()
            h.initialize(scope, tmcc_id=55 if scope != CommandScope.ENGINE else 1234)
            assert h.is_comp_data_record is True
            assert isinstance(h.comp_data, expected_cls)

    def test_comp_data_encode_rpm_labor_and_pkg(self):
        # Encode rpm/labor -> combined byte and into UpdatePkg
        val = CompData.encode_rpm_labor(rpm=5, labor=14)
        assert isinstance(val, int)
        pkg = CompData.rpm_labor_to_pkg(rpm=5, labor=14)
        assert isinstance(pkg, UpdatePkg)
        # data_bytes should be 1 byte representing encoded value
        assert len(pkg.data_bytes) == 1
        assert pkg.length == 1

    def test_missing_attribute_raises_attribute_error(self):
        buf = b"\xff" * PdiReq.scope_record_length(CommandScope.ENGINE)
        eng = EngineData(buf, tmcc_id=1000)
        with pytest.raises(AttributeError):
            _ = eng.this_attribute_does_not_exist  # noqa: B018



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

from src.pytrain.pdi.constants import PDI_EOP, PDI_SOP, PdiCommand, Stm2Action
from src.pytrain.pdi.stm2_req import Stm2Req
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum


def test_config_set_as_bytes_contains_fields():
    """
    CONFIG with SET should serialize tmcc_id, debug, two zero bytes, and mode.
    """
    req = Stm2Req(
        25,
        PdiCommand.STM2_SET,
        Stm2Action.CONFIG,
        mode=0,
        debug=7,
    )

    # Human payload shape
    s = req.payload
    assert isinstance(s, str)
    # Be lenient about exact formatting; check key tokens
    assert "Mode:" in s and "Debug:" in s

    bs = req.as_bytes
    assert isinstance(bs, (bytes, bytearray))
    assert bs[0] == PDI_SOP
    assert bs[-1] == PDI_EOP

    # Core markers
    assert PdiCommand.STM2_SET.as_bytes in bs
    assert Stm2Action.CONFIG.as_bytes in bs

    # The CONFIG payload for SET should include:
    #   tmcc_id, debug, 0x00, 0x00, mode
    expected_tail = bytes([req.tmcc_id, 7, 0x00, 0x00, 0x00])
    assert expected_tail in bs


def test_control1_set_payload_and_state_byte_out():
    """
    CONTROL1 with SET should encode the switch state byte and payload should include 'OUT'.
    """
    req = Stm2Req(10, PdiCommand.STM2_SET, Stm2Action.CONTROL1, state=TMCC1SwitchCommandEnum.OUT)

    s = req.payload
    assert isinstance(s, str)
    assert "OUT" in s  # enum string should contain OUT

    bs = req.as_bytes
    assert bs[0] == PDI_SOP and bs[-1] == PDI_EOP
    assert PdiCommand.STM2_SET.as_bytes in bs
    assert Stm2Action.CONTROL1.as_bytes in bs
    # State byte must appear in the packet; OUT -> 1
    assert (1).to_bytes(1, "big") in bs


def test_control1_set_payload_and_state_byte_thru():
    """
    CONTROL1 with SET should encode the switch state byte and payload should include 'THRU'.
    """
    req = Stm2Req(11, PdiCommand.STM2_SET, Stm2Action.CONTROL1, state=TMCC1SwitchCommandEnum.THRU)

    s = req.payload
    assert isinstance(s, str)
    assert "THRU" in s

    bs = req.as_bytes
    assert bs[0] == PDI_SOP and bs[-1] == PDI_EOP
    assert PdiCommand.STM2_SET.as_bytes in bs
    assert Stm2Action.CONTROL1.as_bytes in bs
    # THRU -> 0
    assert (0).to_bytes(1, "big") in bs


def test_identify_set_includes_ident_byte():
    """
    IDENTIFY with SET should serialize the ident byte.
    """
    req = Stm2Req(9, PdiCommand.STM2_SET, Stm2Action.IDENTIFY, ident=0x55)
    bs = req.as_bytes
    assert bs[0] == PDI_SOP and bs[-1] == PDI_EOP
    assert PdiCommand.STM2_SET.as_bytes in bs
    assert Stm2Action.IDENTIFY.as_bytes in bs
    assert (0x55).to_bytes(1, "big") in bs


def test_get_requests_have_minimal_payload_for_config():
    """
    GET requests should not include the additional CONFIG body bytes.
    """
    # CONFIG + GET should not append the CONFIG body (tmcc_id/debug/zeros/mode).
    req = Stm2Req(19, PdiCommand.STM2_GET, Stm2Action.CONFIG, mode=1, debug=2)
    bs = req.as_bytes
    assert bs[0] == PDI_SOP and bs[-1] == PDI_EOP
    assert PdiCommand.STM2_GET.as_bytes in bs
    assert Stm2Action.CONFIG.as_bytes in bs
    # Ensure the characteristic CONFIG body pattern is absent
    unexpected_tail = bytes([req.tmcc_id, 2, 0x00, 0x00, 0x01])
    assert unexpected_tail not in bs


def test_is_thru_is_out_properties_match_state():
    # OUT
    req_out = Stm2Req(3, PdiCommand.STM2_SET, Stm2Action.CONTROL1, state=TMCC1SwitchCommandEnum.OUT)
    assert req_out.is_out is True
    assert req_out.is_thru is False

    # THRU
    req_thru = Stm2Req(4, PdiCommand.STM2_SET, Stm2Action.CONTROL1, state=TMCC1SwitchCommandEnum.THRU)
    assert req_thru.is_thru is True
    assert req_thru.is_out is False



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

# test/utils/argument_parser_test.py
import argparse

import pytest

from src.pytrain.utils.argument_parser import (
    IntRange,
    PyTrainArgumentParser,
    StripPrefixesHelpFormatter,
)


class TestPyTrainArgumentParser:
    def test_error_raises_without_exit_and_sets_message(self):
        parser = PyTrainArgumentParser(add_help=False)
        parser.clear_exit_on_error()
        with pytest.raises(argparse.ArgumentError) as ei:
            parser.error("boom!")
        assert parser.error_message == "boom!"
        # argparse.ArgumentError stores the message in .message
        assert getattr(ei.value, "message", "") == "boom!"

    def test_exit_raises_without_exit_and_sets_message(self):
        parser = PyTrainArgumentParser(add_help=False)
        parser.clear_exit_on_error()
        with pytest.raises(argparse.ArgumentError) as ei:
            parser.exit(2, "bye!")
        assert parser.error_message == "bye!"
        assert getattr(ei.value, "message", "") == "bye!"

    def test_validate_args_success_and_failure(self):
        parser = PyTrainArgumentParser(add_help=False)
        parser.add_argument("-n", type=int, required=True)

        # Success
        args, msg = parser.validate_args(["-n", "5"])
        assert msg is None
        assert args.n == 5

        # Failure (missing required)
        args, msg = parser.validate_args([])
        assert args == []
        assert isinstance(msg, str)
        assert "required" in msg.lower()

    def test_remove_args_makes_option_unrecognized(self):
        parser = PyTrainArgumentParser(add_help=False)
        parser.add_argument("-foo", type=int, dest="foo")

        # Ensure it parses before removal
        args, msg = parser.validate_args(["-foo", "1"])
        assert msg is None
        assert args.foo == 1

        # Remove and verify it's now unrecognized
        parser.remove_args(["-foo"])
        args, msg = parser.validate_args(["-foo", "1"])
        assert args == []
        assert isinstance(msg, str)
        assert "unrecognized" in msg.lower()

    def test_is_exit_on_error_inherits_from_parent(self):
        parent = PyTrainArgumentParser(add_help=False)
        parent.clear_exit_on_error()
        child = PyTrainArgumentParser(add_help=False, parent=parent)

        # Child should reflect parent's setting
        assert child.is_exit_on_error is False

        # And child.error should raise, not exit
        with pytest.raises(argparse.ArgumentError):
            child.error("child error")


class TestStripPrefixesHelpFormatter:
    def test_usage_strips_hyphens_from_options(self):
        parser = PyTrainArgumentParser(add_help=False, formatter_class=StripPrefixesHelpFormatter)
        parser.add_argument("-ab", action="store_true", help="ab flag")
        parser.add_argument("-cd", action="store_true", help="cd flag")

        usage = parser.format_usage()
        # Expect bare option names without '-' due to custom formatter
        assert "ab" in usage
        assert "cd" in usage
        assert "-ab" not in usage
        assert "-cd" not in usage


class TestIntRange:
    def test_no_bounds_accepts_ints_and_rejects_non_int(self):
        t = IntRange()
        assert t("12") == 12
        with pytest.raises(argparse.ArgumentTypeError) as ei:
            t("x")
        assert "integer" in str(ei.value).lower()

    def test_min_and_max_bounds(self):
        t = IntRange(1, 5)
        assert t("1") == 1
        assert t("5") == 5
        with pytest.raises(argparse.ArgumentTypeError) as ei_low:
            t("0")
        assert "[1 - 5]" in str(ei_low.value)
        with pytest.raises(argparse.ArgumentTypeError) as ei_high:
            t("6")
        assert "[1 - 5]" in str(ei_high.value)

    def test_min_only(self):
        t = IntRange(imin=3)
        assert t("3") == 3
        with pytest.raises(argparse.ArgumentTypeError) as ei:
            t("2")
        assert ">=" in str(ei.value)

    def test_max_only(self):
        t = IntRange(imax=7)
        assert t("7") == 7
        with pytest.raises(argparse.ArgumentTypeError) as ei:
            t("8")
        assert "<=" in str(ei.value)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL

import time
from typing import Any, List

import pytest

from src.pytrain.comm.command_listener import CommandListener
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.multibyte.multibyte_constants import TMCC2EffectsControl as Effects
from src.pytrain.protocol.tmcc1.tmcc1_constants import (
    TMCC1HaltCommandEnum as Halt1,
)
from src.pytrain.protocol.tmcc1.tmcc1_constants import (
    TMCC1SyncCommandEnum,
)
from src.pytrain.protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum as Engine2, TMCC2_MEDIUM_SPEED


class DummyDispatcher:
    def __init__(self):
        self.offered: List[tuple[Any, bool]] = []

    def offer(self, req, from_pdi: bool = False):
        self.offered.append((req, from_pdi))

    # Stubs to satisfy CommandListener API calls
    def shutdown(self): ...
    def subscribe(self, *args, **kwargs): ...
    def unsubscribe(self, *args, **kwargs): ...
    def subscribe_any(self, *args, **kwargs): ...
    def unsubscribe_any(self, *args, **kwargs): ...


def wait_for(predicate, timeout=1.0):
    start = time.time()
    while time.time() - start < timeout:
        if predicate():
            return True
        time.sleep(0.01)
    return False


# noinspection PyProtectedMember
@pytest.fixture(autouse=True)
def reset_listener_singleton(monkeypatch):
    # Do not stub run here; we want the real decoding loop
    # Avoid any startup sync activity
    monkeypatch.setattr(CommandListener, "sync_state", lambda self: None)
    CommandListener._instance = None
    yield
    try:
        if CommandListener._instance is not None:
            CommandListener._instance.shutdown()
    finally:
        CommandListener._instance = None


def build_listener_with_dummy_dispatcher(monkeypatch) -> tuple[CommandListener, DummyDispatcher]:
    dd = DummyDispatcher()
    monkeypatch.setattr(
        "src.pytrain.comm.command_listener.CommandDispatcher.build",
        lambda *a, **k: dd,
    )
    # Important: prevent SerialReader creation by disabling ser2_receiver
    listener = CommandListener(ser2_receiver=False, base3_receiver=False)
    return listener, dd


def test_decode_tmcc1_three_byte_command(monkeypatch):
    listener, dd = build_listener_with_dummy_dispatcher(monkeypatch)

    halt_req = CommandReq.build(Halt1.HALT)
    listener.offer(halt_req.as_bytes)

    assert wait_for(lambda: len(dd.offered) >= 1)
    offered_cmd = dd.offered[0][0]
    assert isinstance(offered_cmd, CommandReq)
    assert offered_cmd.command == Halt1.HALT
    assert offered_cmd.is_halt is True


def test_decode_tmcc2_three_byte_engine_command(monkeypatch):
    listener, dd = build_listener_with_dummy_dispatcher(monkeypatch)

    ring = CommandReq.build(Engine2.RING_BELL, 7)
    listener.offer(ring.as_bytes)

    assert wait_for(lambda: len(dd.offered) >= 1)
    offered_cmd = dd.offered[0][0]
    assert isinstance(offered_cmd, CommandReq)
    assert offered_cmd.command == Engine2.RING_BELL
    assert offered_cmd.address == 7


def test_decode_tmcc2_four_digit_address_command_sets_tmcc4(monkeypatch):
    listener, dd = build_listener_with_dummy_dispatcher(monkeypatch)

    # Build a command with a 4-digit engine id; listener should append the 4 ASCII digits
    cmd_4d = CommandReq.build(Engine2.SPEED_MEDIUM, 1234)
    listener.offer(cmd_4d.as_bytes)

    assert wait_for(lambda: len(dd.offered) >= 1)
    offered_cmd = dd.offered[0][0]
    assert isinstance(offered_cmd, CommandReq)
    assert offered_cmd.command == Engine2.ABSOLUTE_SPEED
    assert offered_cmd.data == TMCC2_MEDIUM_SPEED
    assert offered_cmd.address == 1234  # Decoded as TMCC4 (4-digit) address


def test_decode_tmcc2_multibyte_parameter_command(monkeypatch):
    listener, dd = build_listener_with_dummy_dispatcher(monkeypatch)

    # Effects multibyte command (encodes as >= 9 bytes with multibyte markers)
    # Choose a simple, well-defined effect
    mb_req = CommandReq.build(Effects.SMOKE_HIGH, 22)
    listener.offer(mb_req.as_bytes)

    assert wait_for(lambda: len(dd.offered) >= 1)
    offered_cmd = dd.offered[0][0]
    # MultiByteReq is a subclass of CommandReq; we only validate essential fields here
    assert isinstance(offered_cmd, CommandReq)
    assert offered_cmd.command == Effects.SMOKE_HIGH
    assert offered_cmd.address == 22


def test_decode_admin_sync_command_restart_flows_through_deque(monkeypatch):
    listener, dd = build_listener_with_dummy_dispatcher(monkeypatch)

    restart = CommandReq(TMCC1SyncCommandEnum.RESTART)
    # Only SYNC_BEGIN/SYNC_COMPLETE are handled specially by offer(); others must pass through run()
    listener.offer(restart.as_bytes)

    assert wait_for(lambda: len(dd.offered) >= 1)
    offered_cmd = dd.offered[0][0]
    assert isinstance(offered_cmd, CommandReq)
    assert offered_cmd.command == TMCC1SyncCommandEnum.RESTART



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import uuid

import pytest

from src.pytrain.comm.enqueue_proxy_requests import EnqueueHandler, EnqueueProxyRequests
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import DEFAULT_SERVER_PORT
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum


class DummyBuffer:
    def __init__(self):
        self.enqueued = []
        self.session_id = 123
        self.base3_address = None

    def enqueue_command(self, data: bytes):
        self.enqueued.append(data)


@pytest.fixture(autouse=True)
def reset_singleton(monkeypatch):
    # Prevent the background server thread from creating a real TCP server
    monkeypatch.setattr(EnqueueProxyRequests, "run", lambda self: None)
    # Reset singleton before each test
    EnqueueProxyRequests._instance = None
    yield
    # Reset singleton after each test
    EnqueueProxyRequests._instance = None


def test_build_request_register_with_uuid_and_version():
    port = 12345
    client_id = uuid.uuid4()
    version = (1, 2, 3)
    base = CommandReq(TMCC1SyncCommandEnum.REGISTER).as_bytes

    built = EnqueueProxyRequests.register_request(port, client_id, version)

    assert built.startswith(base)
    # After first 3 bytes, expect: port(2) + uuid(16) + version(3)
    payload = built[3:]
    assert int.from_bytes(payload[:2], "big") == (port & 0xFFFF)
    assert payload[2:18] == client_id.bytes
    assert payload[18:21] == bytes(version)


def test_build_request_disconnect_with_uuid_no_version():
    port = 43210
    client_id = uuid.uuid4()
    base = CommandReq(TMCC1SyncCommandEnum.DISCONNECT).as_bytes

    built = EnqueueProxyRequests.disconnect_request(port, client_id)

    assert built.startswith(base)
    payload = built[3:]
    assert int.from_bytes(payload[:2], "big") == (port & 0xFFFF)
    assert payload[2:18] == client_id.bytes
    # No version afterward
    assert len(payload) == 2 + 16


def test_build_request_sync_state_minimal_defaults():
    # No client_id provided, default port
    base = CommandReq(TMCC1SyncCommandEnum.SYNC_REQUEST).as_bytes

    built = EnqueueProxyRequests.sync_state_request()

    assert built.startswith(base)
    payload = built[3:]
    # default port
    assert int.from_bytes(payload, "big") == (DEFAULT_SERVER_PORT & 0xFFFF)


@pytest.mark.parametrize(
    "cmd_enum",
    [
        TMCC1SyncCommandEnum.QUIT,
        TMCC1SyncCommandEnum.REBOOT,
        TMCC1SyncCommandEnum.RESTART,
        TMCC1SyncCommandEnum.SHUTDOWN,
        TMCC1SyncCommandEnum.UPDATE,
        TMCC1SyncCommandEnum.UPGRADE,
        TMCC1SyncCommandEnum.RESYNC,
        TMCC1SyncCommandEnum.KEEP_ALIVE,
    ],
)
def test_extract_addendum_variants_port_uuid_and_version(cmd_enum):
    port = 25000
    client_id = uuid.uuid4()
    version = (9, 8, 7)

    base = CommandReq(cmd_enum).as_bytes
    # emulate "register-like" framing: 3 bytes header + port(2) + uuid(16) + version(3)
    byte_stream = base + (port & 0xFFFF).to_bytes(2, "big") + client_id.bytes + bytes(version)

    ip, dec_port, dec_uuid, dec_version = EnqueueHandler.extract_addendum(byte_stream)

    assert ip is None
    assert dec_port == port
    assert dec_uuid == client_id
    assert dec_version == version


def test_extract_addendum_port_and_uuid_only():
    port = 11001
    client_id = uuid.uuid4()
    base = CommandReq(TMCC1SyncCommandEnum.DISCONNECT).as_bytes
    byte_stream = base + (port & 0xFFFF).to_bytes(2, "big") + client_id.bytes

    ip, dec_port, dec_uuid, dec_version = EnqueueHandler.extract_addendum(byte_stream)

    assert ip is None
    assert dec_port == port
    assert dec_uuid == client_id
    assert dec_version is None


def test_extract_addendum_ip_port_text():
    ip_text = "192.168.1.77"
    port = 5678
    base = CommandReq(TMCC1SyncCommandEnum.RESTART).as_bytes
    byte_stream = base + f"{ip_text}:{port}".encode("utf-8")

    ip, dec_port, dec_uuid, dec_version = EnqueueHandler.extract_addendum(byte_stream)

    assert ip == ip_text
    assert dec_port == port
    assert dec_uuid is None
    assert dec_version is None


def test_extract_addendum_port_only_two_bytes():
    port = 60001
    base = CommandReq(TMCC1SyncCommandEnum.KEEP_ALIVE).as_bytes
    byte_stream = base + (port & 0xFFFF).to_bytes(2, "big")

    ip, dec_port, dec_uuid, dec_version = EnqueueHandler.extract_addendum(byte_stream)

    assert ip is None
    assert dec_port == (port & 0xFFFF)  # truncated to 16 bits per implementation
    assert dec_uuid is None
    assert dec_version is None


# noinspection PyTypeChecker
def test_client_session_tracking_connect_disconnect_and_replace():
    buf = DummyBuffer()
    proxy = EnqueueProxyRequests(buf, server_port=0)  # run() patched to no-op; thread returns immediately

    ip = "10.0.0.5"
    port = 12345
    cid1 = uuid.uuid4()
    cid2 = uuid.uuid4()

    # Initial connect
    proxy.client_connect(ip, port, cid1)
    assert proxy.is_client(ip, port, cid1) is True
    assert (ip, port) in proxy.client_sessions

    # Connect with same (ip,port) but different client_id should purge prior session
    proxy.client_connect(ip, port, cid2)
    assert proxy.is_client(ip, port, cid1) is False
    assert proxy.is_client(ip, port, cid2) is True
    assert (ip, port) in proxy.client_sessions
    # Only one tuple for (ip,port) across sessions
    assert {(c[0], c[1]) for c in proxy.client_sessions} == {(ip, port)}

    # KEEP_ALIVE refresh doesn't error
    proxy.client_alive(ip, port, cid2)

    # Disconnect
    proxy.client_disconnect(ip, port, cid2)
    assert proxy.is_client(ip, port, cid2) is False
    assert (ip, port) not in proxy.client_sessions


# noinspection PyTypeChecker
def test_enqueue_request_forwards_to_buffer():
    buf = DummyBuffer()
    proxy = EnqueueProxyRequests(buf, server_port=0)

    data = b"\xfe\xf0\x01\x02"
    proxy.enqueue_request(data)

    assert buf.enqueued == [data]
